"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from collections import namedtuple
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from typing import TypeVar, Callable, ClassVar

from gs_quant.backtests.backtest_utils import *
from gs_quant.backtests.backtest_objects import ConstantTransactionModel, TransactionModel
from gs_quant.base import Priceable, static_field
from gs_quant.common import RiskMeasure
from gs_quant.json_convertors import decode_named_instrument, dc_decode, encode_named_instrument, decode_date_or_str
from gs_quant.json_convertors_common import decode_risk_measure, encode_risk_measure
from gs_quant.markets.portfolio import Portfolio
from gs_quant.markets.securities import *
from gs_quant.risk.transform import Transformer
from gs_quant.target.backtests import BacktestTradingQuantityType

action_count = 1


Duration = Union[str, dt.date, dt.timedelta, CustomDuration]


def default_transaction_cost():
    return ConstantTransactionModel(0)


class ScalingActionType(Enum):
    risk_measure = 'risk_measure'
    size = 'size'
    NAV = 'NAV'


@dataclass_json
@dataclass
class Action(object):
    _needs_scaling = False
    _calc_type = CalcType.simple
    _risk = None
    _transaction_cost = ConstantTransactionModel(0)
    name = None
    __sub_classes: ClassVar[List[type]] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Action.__sub_classes.append(cls)

    @staticmethod
    def sub_classes():
        return tuple(Action.__sub_classes)

    def __post_init__(self):
        self.set_name(self.name)

    @property
    def calc_type(self):
        return self._calc_type

    @property
    def risk(self):
        return self._risk

    def set_name(self, name: str):
        global action_count
        if self.name is None:
            self.name = 'Action{}'.format(action_count)
            action_count += 1

    @property
    def transaction_cost(self):
        return self._transaction_cost

    @transaction_cost.setter
    def transaction_cost(self, value):
        self._transaction_cost = value


TAction = TypeVar('TAction', bound='Action')


@dataclass_json
@dataclass
class AddTradeAction(Action):

    """
    create an action which adds a trade when triggered.  The trades are resolved on the trigger date (state) and
    last until the trade_duration if specified or for all future dates if not.
    :param priceables: a priceable or a list of pricables.
    :param trade_duration: an instrument attribute eg. 'expiration_date' or a date or a tenor or timedelta
                           if left as None the
                           trade will be added for all future dates
                           can also specify 'next schedule' in order to exit at the next periodic trigger date
    :param name: optional additional name to the priceable name
    :param transaction_cost: optional a cash amount paid for each transaction, paid on both enter and exit
    """

    priceables: Union[Instrument, Iterable[Instrument]] = field(default=None,
                                                                metadata=config(decoder=decode_named_instrument,
                                                                                encoder=encode_named_instrument))
    trade_duration: Duration = field(default=None,  # de/encoder doesn't handle timedelta
                                     metadata=config(decoder=decode_date_or_str))
    name: str = None
    transaction_cost: TransactionModel = field(default_factory=default_transaction_cost,
                                               metadata=config(decoder=dc_decode(ConstantTransactionModel)))
    holiday_calendar: Iterable[dt.date] = None
    class_type: str = static_field('add_trade_action')

    def __post_init__(self):
        super().__post_init__()
        self._dated_priceables = {}
        named_priceables = []
        for i, p in enumerate(make_list(self.priceables)):
            if p.name is None:
                named_priceables.append(p.clone(name=f'{self.name}_Priceable{i}'))
            elif p.name.startswith(self.name):
                named_priceables.append(p)
            else:
                named_priceables.append(p.clone(name=f'{self.name}_{p.name}'))
        self.priceables = named_priceables
        if self.transaction_cost is None:
            self.transaction_cost = ConstantTransactionModel(0)

    def set_dated_priceables(self, state, priceables):
        self._dated_priceables[state] = make_list(priceables)

    @property
    def dated_priceables(self):
        return self._dated_priceables


AddTradeActionInfo = namedtuple('AddTradeActionInfo', ['scaling', 'next_schedule'])
EnterPositionQuantityScaledActionInfo = namedtuple('EnterPositionQuantityScaledActionInfo', 'not_applicable')
HedgeActionInfo = namedtuple('HedgeActionInfo', 'next_schedule')
ExitTradeActionInfo = namedtuple('ExitTradeActionInfo', 'not_applicable')
RebalanceActionInfo = namedtuple('RebalanceActionInfo', 'not_applicable')
AddScaledTradeActionInfo = namedtuple('AddScaledActionInfo', 'next_schedule')


@dataclass_json
@dataclass
class AddScaledTradeAction(Action):

    """
    create an action which adds a trade when triggered.  The trade is scaled by a measure or trade property.
    The trades are resolved on the trigger date (state) and last until the trade_duration if specified or for
    all future dates if not.
    :param priceables: a priceable or a list of pricables.
    :param trade_duration: an instrument attribute eg. 'expiration_date' or a date or a tenor or timedelta
                           if left as None the
                           trade will be added for all future dates
                           can also specify 'next schedule' in order to exit at the next periodic trigger date
    :param name: optional additional name to the priceable name
    :param scaling_type: the type of scaling we are doing
    :param scaling_risk: if the scaling type is a measure then this is the definition of the measure
    :param scaling_level: the level of scaling to be done
    :param transaction_cost: optional a cash amount paid for each transaction, paid on both enter and exit
    """
    priceables: Union[Priceable, Iterable[Priceable]] = field(default=None,
                                                              metadata=config(decoder=decode_named_instrument,
                                                                              encoder=encode_named_instrument))
    trade_duration: Duration = field(default=None,  # de/encoder doesn't handle timedelta
                                     metadata=config(decoder=decode_date_or_str))
    name: str = None
    scaling_type: ScalingActionType = ScalingActionType.size
    scaling_risk: RiskMeasure = None
    scaling_level: Union[float, int] = 1
    transaction_cost: TransactionModel = field(default_factory=default_transaction_cost,
                                               metadata=config(decoder=dc_decode(ConstantTransactionModel)))
    holiday_calendar: Iterable[dt.date] = None
    class_type: str = static_field('add_scaled_trade_action')

    def __post_init__(self):
        super().__post_init__()
        named_priceables = []
        for i, p in enumerate(make_list(self.priceables)):
            if p.name is None:
                named_priceables.append(p.clone(name=f'{self.name}_Priceable{i}'))
            elif p.name.startswith(self.name):
                named_priceables.append(p)
            else:
                named_priceables.append(p.clone(name=f'{self.name}_{p.name}'))
        self.priceables = named_priceables


@dataclass_json
@dataclass
class EnterPositionQuantityScaledAction(Action):
    """
        create an action which enters trades when triggered.  The trades are executed with specified quantity and
        last until the trade_duration if specified, or for all future dates if not.
        :param priceables: a priceable or a list of pricables.
        :param trade_duration: an instrument attribute eg. 'expiration_date' or a date or a tenor if left as None the
                               trade will be added for all future dates
        :param name: optional additional name to the priceable name
        :param trade_quantity: the amount, in units of trade_quantity_type to be traded
        :param trade_quantity_type: the quantity type used to scale trade. eg. quantity for units, notional for
                                    underlier notional
    """
    priceables: Union[Priceable, Iterable[Priceable]] = field(default=None,
                                                              metadata=config(decoder=decode_named_instrument,
                                                                              encoder=encode_named_instrument))
    trade_duration: Duration = field(default=None,  # de/encoder doesn't handle timedelta
                                     metadata=config(decoder=decode_date_or_str))
    name: str = None
    trade_quantity: float = 1
    trade_quantity_type: BacktestTradingQuantityType = BacktestTradingQuantityType.quantity
    class_type: str = static_field('enter_position_quantity_scaled_action')

    def __post_init__(self):
        super().__post_init__()
        named_priceables = []
        for i, p in enumerate(make_list(self.priceables)):
            if p.name is None:
                named_priceables.append(p.clone(name=f'{self.name}_Priceable{i}'))
            elif p.name.startswith(self.name):
                named_priceables.append(p)
            else:
                named_priceables.append(p.clone(name=f'{self.name}_{p.name}'))
        self.priceables = named_priceables


@dataclass_json
@dataclass
class ExitPositionAction(Action):
    name: str = None
    class_type: str = 'exit_position_action'


@dataclass_json
@dataclass
class ExitTradeAction(Action):
    priceable_names: Union[str, Iterable[str]] = None
    name: str = None
    class_type: str = static_field('exit_trade_action')

    def __post_init__(self):
        super().__post_init__()
        self.priceables_names = make_list(self.priceable_names)


@dataclass_json
@dataclass
class ExitAllPositionsAction(ExitTradeAction):
    """
    Fully exit all held positions
    """
    class_type: str = static_field('exit_all_positions_action')

    def __post_init__(self):
        super().__post_init__()
        self._calc_type = CalcType.path_dependent


@dataclass_json
@dataclass
class HedgeAction(Action):

    """
    create an action which adds a hedge trade when triggered.  This trade will be scaled to hedge the risk
    specified.  The trades are resolved on the trigger date (state) and
    last until the trade_duration if specified or for all future dates if not.
    :param risk: a risk measure which should be hedged
    :param priceables: a priceable or a list of pricables these should have sensitivity to the risk.
    :param trade_duration: an instrument attribute eg. 'expiration_date' or a date or a tenor or timedelta
                           if left as None the
                           trade will be added for all future dates
                           can also specify 'next schedule' in order to exit at the next periodic trigger date
    :param name: optional additional name to the priceable name
    :param transaction_cost: optional a transaction cost model, paid on both enter and exit
    :param risk_transformation: optional a Transformer which will be applied to the raw risk numbers before hedging
    :param holiday_calendar: optional an iterable list of holiday dates
    """

    risk: RiskMeasure = field(default=None, metadata=config(decoder=decode_risk_measure,
                                                            encoder=encode_risk_measure))
    priceables: Optional[Priceable] = field(default=None, metadata=config(decoder=decode_named_instrument,
                                                                          encoder=encode_named_instrument))
    trade_duration: Duration = field(default=None,  # de/encoder doesn't handle timedelta
                                     metadata=config(decoder=decode_date_or_str))
    name: str = None
    csa_term: str = None
    scaling_parameter: str = 'notional_amount'
    transaction_cost: TransactionModel = field(default_factory=default_transaction_cost,
                                               metadata=config(decoder=dc_decode(ConstantTransactionModel)))
    risk_transformation: Transformer = None
    holiday_calendar: Iterable[dt.date] = None
    class_type: str = static_field('hedge_action')

    def __post_init__(self):
        super().__post_init__()
        self._calc_type = CalcType.semi_path_dependent
        if isinstance(self.priceables, Portfolio):
            named_priceables = []
            for i, priceable in enumerate(self.priceables):
                if priceable.name is None:
                    named_priceables.append(priceable.clone(name=f'{self.name}_Priceable{i}'))
                elif priceable.name.startswith(self.name):
                    named_priceables.append(priceable)
                else:
                    named_priceables.append(priceable.clone(name=f'{self.name}_{priceable.name}'))
            named_priceable = Portfolio(named_priceables)
        elif isinstance(self.priceables, Priceable):
            if self.priceables.name is None:
                named_priceable = self.priceables.clone(name=f'{self.name}_Priceable0')
            elif self.priceable.name.startswith(self.name):
                named_priceable = self.priceable
            else:
                named_priceable = self.priceables.clone(name=f'{self.name}_{self.priceables.name}')
        else:
            raise RuntimeError('hedge action only accepts one trade or one portfolio')
        self.priceables = named_priceable

    @property
    def priceable(self):
        return self.priceables


@dataclass_json
@dataclass
class RebalanceAction(Action):
    priceable: Priceable = field(default=None, metadata=config(decoder=decode_named_instrument,
                                                               encoder=encode_named_instrument))
    size_parameter: Union[str, float] = None
    method: Callable = None
    transaction_cost: TransactionModel = field(default_factory=default_transaction_cost,
                                               metadata=config(decoder=dc_decode(ConstantTransactionModel)))
    name: str = None

    def __post_init__(self):
        super().__post_init__()
        self._calc_type = CalcType.path_dependent
        if self.priceable.unresolved is None:
            raise ValueError("Please specify a resolved priceable to rebalance.")
        if self.priceable is not None:
            if self.priceable.name is None:
                self.priceable = self.priceable.clone(name=f'{self.name}_Priceable0')
            else:
                self.priceable = self.priceable.clone(name=f'{self.name}_{self.priceable.name}')



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
import datetime as dt
from enum import Enum
import numpy as np
import pandas as pd
import pytz
from typing import Union, Iterable, ClassVar, List

from gs_quant.backtests.core import ValuationFixingType
from gs_quant.base import field_metadata, static_field
from gs_quant.data import DataFrequency, Dataset
from gs_quant.instrument import Instrument
from gs_quant.json_convertors import decode_pandas_series, encode_pandas_series


class MissingDataStrategy(Enum):
    fill_forward = 'fill_forward'
    interpolate = 'interpolate'
    fail = 'fail'


@dataclass_json
@dataclass
class DataSource:
    __sub_classes: ClassVar[List[type]] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        DataSource.__sub_classes.append(cls)

    @staticmethod
    def sub_classes():
        return tuple(DataSource.__sub_classes)

    def get_data(self, state):
        raise RuntimeError("Implemented by subclass")

    def get_data_range(self, start: Union[dt.date, dt.datetime],
                       end: Union[dt.date, dt.datetime, int]):
        raise RuntimeError("Implemented by subclass")


@dataclass_json
@dataclass
class GsDataSource(DataSource):
    data_set: str
    asset_id: str
    min_date: dt.date = None
    max_date: dt.date = None
    value_header: str = 'rate'
    class_type: str = static_field('gs_data_source')

    def __post_init__(self):
        self.loaded_data = None

    def get_data(self, state: Union[dt.date, dt.datetime] = None):
        if self.loaded_data is None:
            ds = Dataset(self.data_set)
            if self.min_date:
                self.loaded_data = ds.get_data(self.min_date, self.max_date, assetId=(self.asset_id,))
            else:
                return ds.get_data(state, state, assetId=(self.asset_id,))[self.value_header]
        return self.loaded_data[self.value_header].at[pd.to_datetime(state)]

    def get_data_range(self, start: Union[dt.date, dt.datetime], end: Union[dt.date, dt.datetime, int]):
        if self.loaded_data is None:
            ds = Dataset(self.data_set)
            if self.min_date:
                self.loaded_data = ds.get_data(self.min_date, self.max_date, assetId=(self.asset_id,))
            else:
                self.loaded_data = ds.get_data(start, self.max_date, assetId=(self.asset_id,))
        if isinstance(end, int):
            return self.loaded_data.loc[self.loaded_data.index < start].tail(end)
        return self.loaded_data.loc[(start < self.loaded_data.index) & (self.loaded_data.index <= end)]


@dataclass_json
@dataclass
class GenericDataSource(DataSource):
    """
    A data source which holds a pandas series indexed by date or datetime
    :param data_set: a pandas dataframe indexed by date or datetime
    :param missing_data_strategy: MissingDataStrategy which defines behaviour if data is missing, will only take
                                  effect if using get_data, get_data_range has no expectations of the number of
                                  expected data points.
    """
    data_set: pd.Series = field(default=None, metadata=config(decoder=decode_pandas_series,
                                                              encoder=encode_pandas_series))

    missing_data_strategy: MissingDataStrategy = field(default=MissingDataStrategy.fail, metadata=field_metadata)
    class_type: str = static_field('generic_data_source')

    def __eq__(self, other):
        if not isinstance(other, GenericDataSource):
            return False
        return self.missing_data_strategy == other.missing_data_strategy and self.data_set.equals(other.data_set)

    def __post_init__(self):
        self._tz_aware = isinstance(self.data_set.index[0], dt.datetime) and self.data_set.index[0].tzinfo is not None

    def get_data(self, state: Union[dt.date, dt.datetime, Iterable]):
        """
        Get the value of the dataset at a time or date.  If a list of dates or times is provided return list of values
        :param state: a date, datetime or a list of dates or datetimes
        :return: float value
        """
        if isinstance(state, Iterable):
            return [self.get_data(i) for i in state]

        if self._tz_aware and (state.tzinfo is None or state.tzinfo.utcoffset(state) is None):
            state = pytz.utc.localize(state)
        if pd.Timestamp(state) in self.data_set:
            return self.data_set[pd.Timestamp(state)]
        elif state in self.data_set or self.missing_data_strategy == MissingDataStrategy.fail:
            return self.data_set[state]
        else:
            if isinstance(self.data_set.index, pd.DatetimeIndex):
                self.data_set.at[pd.to_datetime(state)] = np.nan
                self.data_set.sort_index(inplace=True)
            else:
                self.data_set.at[state] = np.nan
            self.data_set.sort_index()
            if self.missing_data_strategy == MissingDataStrategy.interpolate:
                self.data_set = self.data_set.interpolate()
            elif self.missing_data_strategy == MissingDataStrategy.fill_forward:
                self.data_set = self.data_set.ffill()
            else:
                raise RuntimeError(f'unrecognised missing data strategy: {str(self.missing_data_strategy)}')
            return self.data_set[pd.to_datetime(state)] if isinstance(self.data_set.index,
                                                                      pd.DatetimeIndex) else self.data_set[state]

    def get_data_range(self, start: Union[dt.date, dt.datetime],
                       end: Union[dt.date, dt.datetime, int]):
        """
        get a range of values from the dataset.
        :param start: a date or datetime
        :param end: a date, datetime or an int.  If an int is provided we return that many data points back from the
                    start date
        :return: pd.Series
        """

        if isinstance(end, int):
            return self.data_set.loc[self.data_set.index < start].tail(end)
        return self.data_set.loc[(start < self.data_set.index) & (self.data_set.index <= end)]


@dataclass_json
@dataclass
class DataManager:
    def __post_init__(self):
        self._data_sources = {}

    def add_data_source(self, series: Union[pd.Series, DataSource], data_freq: DataFrequency,
                        instrument: Instrument, valuation_type: ValuationFixingType):
        if not isinstance(series, DataSource) and not len(series):
            return
        if instrument.name is None:
            raise RuntimeError('Please add a name identify your instrument')
        key = (data_freq, instrument.name, valuation_type)
        if key in self._data_sources:
            raise RuntimeError('A dataset with this frequency instrument name and valuation type already added to '
                               'Data Manager')
        self._data_sources[key] = GenericDataSource(series) if isinstance(series, pd.Series) else series

    def get_data(self, state: Union[dt.date, dt.datetime], instrument: Instrument, valuation_type: ValuationFixingType):
        key = (DataFrequency.REAL_TIME if isinstance(state, dt.datetime) else DataFrequency.DAILY,
               instrument.name.split('_')[-1], valuation_type)
        return self._data_sources[key].get_data(state)

    def get_data_range(self, start: Union[dt.date, dt.datetime],
                       end: Union[dt.date, dt.datetime], instrument: Instrument, valuation_type: ValuationFixingType):
        key = (DataFrequency.REAL_TIME if isinstance(start, dt.datetime) else DataFrequency.DAILY,
               instrument.name.split('_')[-1], valuation_type)
        return self._data_sources[key].get_data_range(start, end)



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt
import numpy as np
import pandas as pd
from numpy import int64
from pandas._testing import assert_series_equal, assert_frame_equal

from gs_quant.api.gs.backtests_xasset.json_encoders.response_datatypes.risk_result_datatype_encoders import \
    encode_series_result, encode_dataframe_result, decode_series_result, decode_dataframe_result


def test_encode_series_result():
    d = {'a': 1, 'b': 2, 'c': 3}
    s = pd.Series(data=d, index=['a', 'b', 'c'], name='test')
    enc = encode_series_result(s)
    assert enc == {'index': ('a', 'b', 'c'), 'name': 'test', 'values': (1, 2, 3)}


def test_encode_dataframe_result():
    df = pd.DataFrame(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), columns=['a', 'b', 'c'])
    enc = encode_dataframe_result(df)
    assert enc == {'index': (0, 1, 2), 'columns': ('a', 'b', 'c'), 'values': ((1, 2, 3), (4, 5, 6), (7, 8, 9))}


def test_decode_series_result():
    enc = {'index': ('2024-06-12', '2024-06-13', '2024-06-14'), 'name': 'test', 'values': (1, 2, 3)}
    s = decode_series_result(enc)
    assert_series_equal(s, pd.Series(data=[1, 2, 3],
                                     index=[dt.date(2024, 6, 12), dt.date(2024, 6, 13), dt.date(2024, 6, 14)],
                                     name='test'))


def test_decode_dataframe_result():
    enc = {'index': ('2024-06-12', '2024-06-13', '2024-06-14'), 'columns': ('a', 'b', 'c'),
           'values': ((1, 2, 3), (4, 5, 6), (7, 8, 9))}
    df = decode_dataframe_result(enc)
    assert_frame_equal(df, pd.DataFrame(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), columns=['a', 'b', 'c'],
                       index=[dt.date(2024, 6, 12), dt.date(2024, 6, 13), dt.date(2024, 6, 14)], dtype=int64))



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import hashlib
import json
import logging
import os
from abc import abstractmethod
from os.path import exists
from pathlib import Path
from typing import List, Dict, Tuple, Any, NamedTuple
from unittest import mock

from gs_quant.errors import MqUninitialisedError
from gs_quant.session import GsSession, Environment
from gs_quant.test.mock_data_test_utils import log_mock_data_event

logger = logging.getLogger("mock_request")


class MockFileKey:

    def __init__(self, request_id):
        self.request_id = request_id.decode('utf-8') if isinstance(request_id, bytes) else request_id
        self.request_hash = self.get_hash(self.request_id)
        self.file_name = f'request{self.request_hash}.json'

    @staticmethod
    def get_hash(request_id: str):
        id_bytes = request_id.encode('utf-8')
        return hashlib.md5(id_bytes).hexdigest()

    def __eq__(self, other):
        return self.file_name == other.file_name and self.request_id == other.request_id and \
            self.request_hash == other.request_hash

    def __hash__(self):
        return hash((self.file_name, self.request_id, self.request_hash))

    def __repr__(self):
        return f"MockFileKey({self.request_hash})"


class MockFileRecord(NamedTuple):
    """a docstring"""
    foo: int
    bar: str
    used_in_tests: set


class MockRequest:
    __looked_at_files = {}
    __cached_test_data = {}
    __saved_files = set()
    api = None
    method = None
    api_method = None

    def __init__(self, mocker, save_files=False, paths=None, application='gs-quant'):
        # do not save tests with save_files = True
        self.save_files = save_files
        self.mocker = mocker
        self.paths = paths
        self.application = application
        if any([attr is None for attr in [self.api, self.method, self.api_method]]):
            raise NotImplementedError('Mock Subclasses must implement api and method properties')

    def __enter__(self):
        if self.save_files:
            GsSession.use(Environment.PROD, None, None, application=self.application)
            self.mocker.patch.object(self.api, self.method, side_effect=self.mock_calc_create_new_files if str(
                self.save_files).casefold() == 'new' else self.mock_calc_check_and_create_files)
        else:
            try:
                _ = GsSession.current
            except MqUninitialisedError:
                from gs_quant.session import OAuth2Session
                OAuth2Session._authenticate = mock.MagicMock(return_value=None)
                GsSession.use(Environment.PROD, 'fake_client_id', 'fake_secret', application=self.application)
            self.mocker.patch.object(self.api, self.method, side_effect=self.mock_calc)

    def full_file_path(self, mock_file_key: MockFileKey):
        return self.build_file_path(self.paths, mock_file_key)

    @staticmethod
    def build_file_path(base_path, mock_file_key: MockFileKey):
        return base_path / 'calc_cache' / mock_file_key.file_name

    def mock_calc(self, *args, **kwargs):
        request_id = self.get_request_id(args, kwargs)
        mfk = MockFileKey(request_id)
        file_path = self.full_file_path(mfk)
        if not exists(file_path):
            logger.error(f'Unable to find mock test date for {self.__class__.__name__}:\n'
                         f'File {file_path} not found. Key: {request_id}')
            raise FileNotFoundError(f'File {mfk.file_name} not found in {self.paths / "calc_cache"}')
        existing_meta, mock_data = self._load_mock_data_from_file(file_path)
        self._record_file_used_in_test(mfk)
        return mock_data

    def _record_file_used_in_test(self, mfk: MockFileKey):
        test_name = self._get_current_test_name()
        file_deets = MockRequest.__looked_at_files.setdefault(self.paths, dict())
        file_deets.setdefault(mfk, (self.__class__.__name__, set()))[1].add(test_name)

    @staticmethod
    def _get_current_test_name():
        return os.environ.get('PYTEST_CURRENT_TEST').split(':')[-1].split(' ')[0]

    @classmethod
    def _load_mock_data_from_file(cls, file_path):
        meta = {}
        data = cls.__cached_test_data.get(file_path)
        if data is None:
            with open(file_path) as json_data:
                data = json.load(json_data)
                cls.__cached_test_data[file_path] = data
        if isinstance(data, dict) and 'mocked_data' in data:
            meta = data
            data = json.loads(data['mocked_data'])
        return meta, data

    @classmethod
    def _make_mock_file_data(cls, mfk: MockFileKey, cls_name, mock_data_str, used_in_tests):
        return {
            'request_id': mfk.request_id,
            'request_hash': mfk.request_hash,
            'type': cls_name,
            'tests': used_in_tests,
            'mocked_data': mock_data_str,
        }

    def create_files(self, request_id, result_json):
        cls_name = self.__class__.__name__
        ct = self._get_current_test_name()
        mfk = MockFileKey(request_id)
        file_path = self.full_file_path(mfk)
        mock_data = self._make_mock_file_data(mfk, cls_name, result_json.decode('utf-8'), [ct])
        with open(file_path, 'w') as file_data:
            MockRequest.__saved_files.add(request_id)
            file_data.write(json.dumps(mock_data, indent=4))
        log_mock_data_event(f'Created {mfk.file_name}.')
        self._record_file_used_in_test(mfk)

    @abstractmethod
    def mock_calc_create_files(self, *args, **kwargs):
        ...

    def mock_calc_check_and_create_files(self, *args, **kwargs):
        # Check if we have already saved this file in this test run, in which case skip re-saving it
        request_id = self.get_request_id(args, kwargs)
        if request_id in MockRequest.__saved_files:
            res = self.mock_calc(*args, **kwargs)
            # This file was remade this session so we need to reindex it here
            self._reindex_single_file(MockFileKey(request_id))
            return res
        return self.mock_calc_create_files(*args, **kwargs)

    def mock_calc_create_new_files(self, *args, **kwargs):
        request_id = self.get_request_id(args, kwargs)
        full_file_path = self.full_file_path(MockFileKey(request_id))
        file_exists = exists(full_file_path)
        if file_exists:
            return self.mock_calc(*args, **kwargs)
        else:
            return self.mock_calc_create_files(*args, **kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @staticmethod
    def get_file_summary() -> Dict[Path, Dict[MockFileKey, Tuple[List[str], Tuple[str, str, str]]]]:
        return MockRequest.__looked_at_files

    @classmethod
    def get_unused_files(cls, log=False) -> Tuple[Dict[str, Any], ...]:
        file_summary = cls.get_file_summary()
        unused_files = []
        for path, files_used in file_summary.items():
            files_names_used = set(m.file_name for m in files_used.keys())
            for test_file in os.listdir(path / 'calc_cache'):
                if test_file.endswith('.json') and test_file not in files_names_used:
                    full_path = path / 'calc_cache' / test_file
                    try:
                        unused_meta, _ = cls._load_mock_data_from_file(full_path)
                    except Exception:
                        unused_meta = {'tests': ['UNKNOWN']}
                    if log:
                        log_mock_data_event(f'Noticed UNUSED {test_file}.')
                    unused_details = {'file': test_file, 'full_path': full_path,
                                      'tests': tuple(unused_meta.get('tests', []))}
                    unused_files.append(unused_details)
        return tuple(unused_files)

    @classmethod
    def remove_unused_files(cls):
        unused_files = cls.get_unused_files()
        for file in unused_files:
            try:
                log_mock_data_event(f'Removing {file["file"]}. As not used.')
                os.remove(file['full_path'])
            except Exception:
                continue

    @classmethod
    def reindex_test_files(cls, report_only=False, log=False) -> Tuple[str, ...]:
        reindex_files = []
        file_summary = cls.get_file_summary()
        for path, files_used in file_summary.items():
            for mock_file_key, (cls_name, now_used) in files_used.items():
                updated = cls._update_single_file_with_new_tests(path, mock_file_key, now_used, report_only, log)
                if updated:
                    reindex_files.append(mock_file_key.file_name)
        return tuple(reindex_files)

    @classmethod
    def _update_single_file_with_new_tests(cls, base_path, mock_file_key, new_tests, report_only, log):
        full_file_path = cls.build_file_path(base_path, mock_file_key)
        existing_meta, data = cls._load_mock_data_from_file(full_file_path)
        existing_used = tuple(existing_meta.get('tests', []))
        sorted_tests = tuple(sorted(new_tests))
        if existing_used == sorted_tests:
            return False
        if report_only:
            if log:
                missing = set(existing_used) - set(sorted_tests)
                extra = set(sorted_tests) - set(existing_used)
                missing_txt = f" {len(missing)} test(s) are listed in the file but weren't used: {missing}." \
                    if len(missing) else ""
                extra_txt = f" {len(extra)} test(s) that accessed this file were NOT listed in the file : {extra}." \
                    if len(extra) else ""
                log_mock_data_event(f'Noticed BAD INDEX {mock_file_key.file_name}. {missing_txt}{extra_txt}')
            return True

        new_data = cls._make_mock_file_data(mock_file_key, existing_meta['type'], json.dumps(data), sorted_tests)
        # update the cached in memory test data with new contents
        cls.__cached_test_data[full_file_path] = new_data
        # Write the updated mock file
        with open(full_file_path, 'w') as out_file:
            out_file.write(json.dumps(new_data, indent=4))
        log_mock_data_event(f'Updated {mock_file_key.file_name}. As {existing_used} != {sorted_tests}')
        return True

    def _reindex_single_file(self, mock_file_key: MockFileKey):
        file_summary = self.get_file_summary()
        by_path = file_summary.get(self.paths, dict())
        mock_file_usage = by_path.get(mock_file_key)
        if mock_file_usage is None:
            return
        test_used = mock_file_usage[1]
        self._update_single_file_with_new_tests(self.paths, mock_file_key, test_used, report_only=False, log=True)

    @staticmethod
    def get_saved_files() -> List[str]:
        return list(MockRequest.__saved_files)

    @abstractmethod
    def get_request_id(self, args, kwargs):
        ...



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import pandas as pd
import re
from typing import Optional, Union, Iterable, Dict, Tuple, Any

from dataclasses_json import config
from dataclasses import MISSING, fields
from dateutil.parser import isoparse


__valid_date_formats = ('%Y-%m-%d',  # '2020-07-28'
                        '%d%b%y',  # '28Jul20'
                        '%d-%b-%y',  # '28-Jul-20'
                        '%d/%m/%Y')  # '28/07/2020

DateOrDateTime = Union[dt.date, dt.datetime]


def encode_date_or_str(value: Optional[Union[str, dt.date]]) -> Optional[str]:
    return value.isoformat() if isinstance(value, dt.date) else value


def decode_optional_date(value: Optional[str]) -> Optional[dt.date]:
    # from dataclasses-json 0.6.5 onwards the global config for type T will be applied to Optional[T]
    # So this decoder would become redundant, to allow any version we simply return if it's already a date
    if value is None or isinstance(value, dt.date):
        return value
    elif isinstance(value, str):
        return dt.datetime.strptime(value, '%Y-%m-%d').date()

    raise ValueError(f'Cannot convert {value} to date')


def decode_optional_time(value: Optional[str]) -> Optional[dt.time]:
    # from dataclasses-json 0.6.5 onwards the global config for type T will be applied to Optional[T]
    # So this decoder would become redundant, to allow any version we simply return if it's already a time
    if value is None or isinstance(value, dt.time):
        return value
    elif isinstance(value, str):
        return dt.time.fromisoformat(value)

    raise ValueError(f'Cannot convert {value} to date')


def encode_optional_time(value: Optional[Union[str, dt.time]]) -> Optional[str]:
    return value.isoformat() if isinstance(value, dt.time) else value


def decode_date_tuple(blob: Tuple[str, ...]):
    return tuple(decode_optional_date(s) for s in blob) if isinstance(blob, (tuple, list)) else None


def encode_date_tuple(values: Tuple[Optional[Union[str, dt.date]], ...]):
    return tuple(encode_date_or_str(value) if isinstance(value, (str, dt.date)) else None for value in values) if \
        values is not None else None


def decode_iso_date_or_datetime(value: Any) -> Union[Tuple[DateOrDateTime, ...], DateOrDateTime]:
    if isinstance(value, (tuple, list)):
        return tuple(decode_iso_date_or_datetime(v) for v in value)
    if isinstance(value, (dt.date, dt.datetime)):
        return value
    elif isinstance(value, str):
        if len(value) == 10:
            return decode_optional_date(value)
        else:
            return optional_from_isodatetime(value)
    raise TypeError(f'Cannot convert {value} to date or datetime')


def optional_from_isodatetime(datetime: Union[str, dt.datetime, None]) -> Optional[dt.datetime]:
    if datetime is None or isinstance(datetime, dt.datetime):
        return datetime
    return dt.datetime.fromisoformat(datetime.replace('Z', ''))


def optional_to_isodatetime(datetime: Optional[dt.datetime]):
    return f'{dt.datetime.isoformat(datetime, timespec="seconds")}Z' if datetime is not None else None


optional_datetime_config = config(encoder=optional_to_isodatetime, decoder=optional_from_isodatetime)
optional_date_config = config(encoder=encode_date_or_str, decoder=decode_optional_date)


def decode_dict_date_key(value):
    return {dt.date.fromisoformat(d): v for d, v in value.items()} if value is not None else None


def decode_dict_dict_date_key(value):
    return {k: {dt.date.fromisoformat(d): v for d, v in val.items()} if val is not None else None
            for k, val in value.items()} if value is not None else None


def decode_dict_date_value(value):
    return {k: dt.date.fromisoformat(d) for k, d in value.items()} if value is not None else None


def decode_datetime_tuple(blob: Tuple[str, ...]):
    return tuple(optional_from_isodatetime(s) for s in blob) if isinstance(blob, (tuple, list)) else None


def decode_date_or_str(value: Union[dt.date, float, str]) -> Optional[Union[dt.date, str]]:
    if value is None or isinstance(value, dt.date):
        return value
    elif isinstance(value, float):
        # Assume it's an Excel date
        if value > 59:
            value -= 1  # Excel leap year bug, 1900 is not a leap year!
        return (dt.datetime(1899, 12, 31) + dt.timedelta(days=value)).date()
    elif isinstance(value, str):
        # Try the supported string date formats
        for fmt in __valid_date_formats:
            try:
                return dt.datetime.strptime(value, fmt).date()
            except ValueError:
                pass

        # Assume it's a tenor
        return value

    raise TypeError(f'Cannot convert {value} to date')


def encode_datetime(value: Optional[dt.datetime]) -> Optional[str]:
    if value is None:
        return value

    try:
        iso_formatted = value.isoformat(timespec='milliseconds')
    except TypeError:
        # Pandas Timestamp objects don't take timespec, will raise TypeError (as of 1.2.4)
        iso_formatted = value.isoformat()

    return iso_formatted if value.tzinfo else iso_formatted + 'Z'  # Make sure to be explict about timezone


def decode_datetime(value: Optional[Union[int, str]]) -> Optional[dt.datetime]:
    if value is None or isinstance(value, dt.datetime):
        return value
    if isinstance(value, int):
        return dt.datetime.fromtimestamp(value / 1000)
    elif isinstance(value, str):
        matcher = re.search('\\.([0-9]*)Z$', value)
        if matcher:
            sub_seconds = matcher.group(1)
            if len(sub_seconds) > 6:
                value = re.sub(matcher.re, '.{}Z'.format(sub_seconds[:6]), value)

        return isoparse(value)

    raise TypeError(f'Cannot convert {value} to datetime')


def decode_float_or_str(value: Optional[Union[float, int, str]]) -> Optional[Union[float, str]]:
    if value is None:
        return value
    elif isinstance(value, float):
        return value
    elif isinstance(value, int):
        return float(value)
    elif isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            # Assume it's a strike or similar, e.g. 'ATM'
            return value

    raise TypeError(f'Cannot convert {value} to float')


def decode_instrument(value: Optional[Dict]):
    from gs_quant.instrument import Instrument
    return Instrument.from_dict(value) if value else None


def decode_named_instrument(value: Optional[Union[Iterable[Dict], dict]]):
    from gs_quant.instrument import Instrument
    if isinstance(value, (list, tuple)):
        return tuple(decode_named_instrument(v) for v in value)
    elif isinstance(value, dict) and 'portfolio_name' in value.keys():
        return decode_named_portfolio(value)
    return Instrument.from_dict(value) if value else None


def decode_named_portfolio(value):
    from gs_quant.markets.portfolio import Portfolio
    return Portfolio([decode_named_instrument(v) for v in value['instruments']],
                     name=value['portfolio_name'])


def encode_named_instrument(obj):
    from gs_quant.markets.portfolio import Portfolio
    if isinstance(obj, (list, tuple)):
        return tuple(encode_named_instrument(o) for o in obj)
    elif isinstance(obj, Portfolio):
        return encode_named_portfolio(obj)
    return obj.as_dict()


def encode_named_portfolio(obj):
    return {'portfolio_name': obj.name,
            'instruments': tuple(encode_named_instrument(o) for o in obj.all_instruments)}


def encode_pandas_series(obj):
    series_dict = pd.Series.to_dict(obj)
    if isinstance(next(iter(series_dict)), (dt.date, dt.datetime)):
        series_dict = {k.isoformat(): v for k, v in series_dict.items()}
    return series_dict


def decode_pandas_series(value: dict):
    dated_dict = {decode_iso_date_or_datetime(k): v for k, v in value.items()}
    return pd.Series(dated_dict)


def decode_quote_report(value: Optional[dict]):
    from gs_quant.quote_reports.core import quote_report_from_dict
    return quote_report_from_dict(value) if value else None


def decode_quote_reports(value: Optional[Iterable[Dict]]):
    from gs_quant.quote_reports.core import quote_reports_from_dicts
    return quote_reports_from_dicts(value) if value else None


def decode_custom_comment(value: Optional[dict]):
    from gs_quant.quote_reports.core import custom_comment_from_dict
    return custom_comment_from_dict(value) if value else None


def decode_custom_comments(value: Optional[Iterable[Dict]]):
    from gs_quant.quote_reports.core import custom_comments_from_dicts
    return custom_comments_from_dicts(value) if value else None


def decode_hedge_type(value: Optional[dict]):
    from gs_quant.quote_reports.core import hedge_type_from_dict
    return hedge_type_from_dict(value) if value else None


def decode_hedge_types(value: Optional[Iterable[Dict]]):
    from gs_quant.quote_reports.core import hedge_type_from_dicts
    return hedge_type_from_dicts(value) if value else None


def encode_dictable(o):
    return o if o is None else o.to_dict()


def encode_named_dictable(o):
    d = encode_dictable(o)
    if d is not None:
        d['type'] = type(o).__name__
    return d


def _get_dc_type(cls, name_field: str, allow_missing: bool):
    type_field = list(filter(lambda f: f.name in (name_field, f'{name_field}_'), fields(cls)))
    if len(type_field) == 0:
        if allow_missing:
            return None
        raise ValueError(f'Class {cls} has no "{name_field}" property')
    def_value = type_field[0].default
    if def_value == MISSING or def_value is None:
        raise ValueError('No default value for "class_type" field on class')
    return def_value


def _value_decoder(type_to_cls_map, explicit_cls=None, str_mapper=None):
    def decode_value(value):
        if value is None or 'null' == value:
            return None
        if isinstance(value, (list, tuple)):
            return tuple(decode_value(v) for v in value)
        if isinstance(value, str):
            return str_mapper(value) if str_mapper is not None else value
        if isinstance(value, (float, int, dt.date)):
            return value
        if not isinstance(value, dict):
            raise TypeError(f'Cannot decode object of type: {type(value)}')
        if explicit_cls is not None:
            return explicit_cls.from_dict(value)
        else:
            if 'class_type' not in value:
                raise ValueError(f'Object has no "class_type" property {value}')
            obj_type = value['class_type']
            if obj_type not in type_to_cls_map:
                raise ValueError(f'No class mapping for object type: "{obj_type}"')
            try:
                return type_to_cls_map[obj_type].from_dict(value)
            except Exception as e:
                raise ValueError(f'Failed to de-serialise {type_to_cls_map[obj_type]} from value {value}') from e

    return decode_value


def dc_decode(*classes, name_field='class_type', allow_missing=False):
    mappings = ((_get_dc_type(cls, name_field, allow_missing), cls) for cls in classes)
    type_to_cls_map = dict((k, v) for k, v in mappings if k is not None)
    return _value_decoder(type_to_cls_map, None)



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import copy
import datetime as dt
import logging
from collections import defaultdict
from datetime import date
from functools import reduce
from itertools import zip_longest
from typing import Union, Iterable, Optional, Dict

from gs_quant import risk
from gs_quant.backtests.action_handler import ActionHandlerBaseFactory, ActionHandler
from gs_quant.backtests.actions import (Action, AddTradeAction, HedgeAction, EnterPositionQuantityScaledAction,
                                        AddTradeActionInfo, HedgeActionInfo, ExitTradeAction, ExitTradeActionInfo,
                                        EnterPositionQuantityScaledActionInfo, RebalanceAction, RebalanceActionInfo,
                                        ExitAllPositionsAction, AddScaledTradeAction, ScalingActionType,
                                        AddScaledTradeActionInfo)
from gs_quant.backtests.backtest_engine import BacktestBaseEngine
from gs_quant.backtests.backtest_objects import BackTest, ScalingPortfolio, CashPayment, Hedge, PnlDefinition
from gs_quant.backtests.backtest_utils import make_list, CalcType, get_final_date
from gs_quant.backtests.strategy import Strategy
from gs_quant.common import AssetClass, Currency, ParameterisedRiskMeasure, RiskMeasure
from gs_quant.context_base import nullcontext
from gs_quant.datetime.relative_date import RelativeDateSchedule
from gs_quant.markets import PricingContext, HistoricalPricingContext
from gs_quant.markets.portfolio import Portfolio
from gs_quant.risk import Price
from gs_quant.risk.results import PortfolioRiskResult
from gs_quant.target.backtests import BacktestTradingQuantityType
from gs_quant.target.measures import ResolvedInstrumentValues
from gs_quant.tracing import Tracer

# priority set to contexts making requests to the pricing API (min. 1 - max. 10)
DEFAULT_REQUEST_PRIORITY = 5


def raiser(ex):
    raise RuntimeError(ex)


logger = logging.getLogger(__name__)


# Action Implementations
class AddTradeActionImpl(ActionHandler):
    def __init__(self, action: AddTradeAction):
        super().__init__(action)

    def _raise_order(self,
                     state: Union[date, Iterable[date]],
                     trigger_info: Optional[Union[AddTradeActionInfo, Iterable[AddTradeActionInfo]]] = None):
        with PricingContext():
            state_list = make_list(state)
            orders = {}
            if trigger_info is None or isinstance(trigger_info, AddTradeActionInfo):
                trigger_info = [trigger_info for _ in range(len(state_list))]
            for s, ti in zip_longest(state_list, trigger_info):
                active_portfolio = self.action.dated_priceables.get(s) or self.action.priceables
                with PricingContext(pricing_date=s):
                    orders[s] = (Portfolio(active_portfolio).resolve(in_place=False), ti)
        final_orders = {}
        for d, p in orders.items():
            new_port = Portfolio([t.clone(name=f'{t.name}_{d}') for t in p[0].result()])
            final_orders[d] = (new_port.scale(None if p[1] is None else p[1].scaling, in_place=False), p[1])

        return final_orders

    def apply_action(self,
                     state: Union[date, Iterable[date]],
                     backtest: BackTest,
                     trigger_info: Optional[Union[AddTradeActionInfo, Iterable[AddTradeActionInfo]]] = None):

        orders = self._raise_order(state, trigger_info)

        # record entry and unwind cashflows
        for create_date, (portfolio, info) in orders.items():
            for inst in portfolio.all_instruments:
                backtest.cash_payments[create_date].append(CashPayment(inst, effective_date=create_date, direction=-1))
                backtest.transaction_costs[create_date] -= self.action.transaction_cost.get_cost(create_date, backtest,
                                                                                                 trigger_info, inst)
                final_date = get_final_date(inst, create_date, self.action.trade_duration, self.action.holiday_calendar,
                                            info)
                backtest.cash_payments[final_date].append(CashPayment(inst, effective_date=final_date))
                backtest.transaction_costs[final_date] -= self.action.transaction_cost.get_cost(final_date,
                                                                                                backtest,
                                                                                                trigger_info, inst)
                backtest_states = (s for s in backtest.states if final_date > s >= create_date)
                for s in backtest_states:
                    backtest.portfolio_dict[s].append(inst)

        return backtest


class AddScaledTradeActionImpl(ActionHandler):
    def __init__(self, action: AddScaledTradeAction):
        super().__init__(action)

    def _nav_scale_orders(self, orders, price_measure, trigger_infos):
        sorted_order_days = sorted(make_list(orders.keys()))
        final_days_orders = {}

        # Populate dict of dates and instruments sold on those dates
        for create_date, portfolio in orders.items():
            info = trigger_infos[create_date]
            for inst in portfolio.all_instruments:
                d = get_final_date(inst, create_date, self.action.trade_duration, self.action.holiday_calendar, info)
                if d not in final_days_orders.keys():
                    final_days_orders[d] = []
                final_days_orders[d].append(inst)

        unscaled_prices_by_day = {}
        unscaled_unwind_prices_by_day = {}
        with PricingContext(is_async=True):
            for day, portfolio in orders.items():
                with PricingContext(pricing_date=day):
                    unscaled_prices_by_day[day] = portfolio.calc(price_measure)
            for unwind_day, unwind_instruments in final_days_orders.items():
                if unwind_day <= dt.date.today():
                    with PricingContext(pricing_date=unwind_day):
                        unscaled_unwind_prices_by_day[unwind_day] = Portfolio(unwind_instruments).calc(price_measure)

        # Start with first_quantity, then only use proceeds from selling instruments
        available_cash = self.action.scaling_level
        scaling_factors_by_inst = {}
        # Go through each order day of the strategy in sorted order
        for idx, cur_day in enumerate(sorted_order_days):
            scale_factor = available_cash / unscaled_prices_by_day[cur_day].aggregate()
            portfolio = orders[cur_day]
            portfolio.scale(scale_factor)
            for inst in portfolio:
                scaling_factors_by_inst[inst] = scale_factor

            available_cash = 0

            if idx + 1 < len(sorted_order_days):
                next_day = sorted_order_days[idx + 1]
            else:
                break

            # Cash received from unwinds is the cash available for the next order
            for d, p in final_days_orders.items():
                # Only consider final days between current order date and the next in an iteration
                if cur_day < d <= next_day:
                    available_cash += sum(unscaled_unwind_prices_by_day[d][inst] * scaling_factors_by_inst[inst] for
                                          inst in p)

    def _scale_order(self, orders, daily_risk, price_measure, trigger_infos):
        if self.action.scaling_type == ScalingActionType.size:
            for _, portfolio in orders.items():
                portfolio.scale(self.action.scaling_level)
        elif self.action.scaling_type == ScalingActionType.NAV:
            self._nav_scale_orders(orders, price_measure, trigger_infos)
        elif self.action.scaling_type == ScalingActionType.risk_measure:
            for day, portfolio in orders.items():
                scaling_factor = self.action.scaling_level / daily_risk[day]
                portfolio.scale(scaling_factor)
        else:
            raise RuntimeError(f'Scaling Type {self.action.scaling_type} not supported by engine')

    def _raise_order(self,
                     state_list: Iterable[date],
                     price_measure: RiskMeasure,
                     trigger_infos: Dict[dt.date, Optional[Union[AddScaledTradeActionInfo,
                                                                 Iterable[AddScaledTradeActionInfo]]]]):
        orders = {}
        order_valuations = (ResolvedInstrumentValues,)
        if self.action.scaling_type == ScalingActionType.risk_measure:
            order_valuations += (self.action.scaling_risk,)
        with PricingContext():
            for s in state_list:
                with PricingContext(pricing_date=s):
                    orders[s] = Portfolio(self.action.priceables).calc(order_valuations)

        final_orders = {}
        for d, res in orders.items():
            new_port = []
            for inst in self.action.priceables:
                new_inst = res[inst]
                if len(order_valuations) > 1:
                    new_inst = new_inst[ResolvedInstrumentValues]
                new_inst.name = f'{new_inst.name}_{d}'
                new_port.append(new_inst)
            final_orders[d] = Portfolio(new_port)
        daily_risk = {d: res[self.action.scaling_risk].aggregate() for d, res in orders.items()} if \
            self.action.scaling_type == ScalingActionType.risk_measure else None

        self._scale_order(final_orders, daily_risk, price_measure, trigger_infos)

        return final_orders

    def apply_action(self,
                     state: Union[date, Iterable[date]],
                     backtest: BackTest,
                     trigger_info: Optional[Union[AddScaledTradeActionInfo,
                                                  Iterable[AddScaledTradeActionInfo]]] = None):

        state_list = make_list(state)
        if trigger_info is None or isinstance(trigger_info, AddScaledTradeActionInfo):
            trigger_info = [trigger_info for _ in range(len(state_list))]
        trigger_infos = dict(zip_longest(state_list, trigger_info))
        orders = self._raise_order(state_list, backtest.price_measure, trigger_infos)

        # record entry and unwind cashflows
        for create_date, portfolio in orders.items():
            info = trigger_infos[create_date]
            for inst in portfolio.all_instruments:
                backtest.cash_payments[create_date].append(CashPayment(inst, effective_date=create_date, direction=-1))
                backtest.transaction_costs[create_date] -= self.action.transaction_cost.get_cost(create_date, backtest,
                                                                                                 trigger_info, inst)
                final_date = get_final_date(inst, create_date, self.action.trade_duration, self.action.holiday_calendar,
                                            info)
                backtest.cash_payments[final_date].append(CashPayment(inst, effective_date=final_date))
                backtest.transaction_costs[final_date] -= self.action.transaction_cost.get_cost(final_date,
                                                                                                backtest,
                                                                                                trigger_info, inst)
                backtest_states = (s for s in backtest.states if final_date > s >= create_date)
                for s in backtest_states:
                    backtest.portfolio_dict[s].append(inst)

        return backtest


class EnterPositionQuantityScaledActionImpl(ActionHandler):
    def __init__(self, action: EnterPositionQuantityScaledAction):
        super().__init__(action)

    @staticmethod
    def _quantity_type_to_risk_measure(quantity_type: BacktestTradingQuantityType):
        map = {BacktestTradingQuantityType.notional: risk.EqSpot,
               BacktestTradingQuantityType.gamma: risk.EqGamma,
               BacktestTradingQuantityType.vega: risk.EqVega}

        if quantity_type not in map:
            raise ValueError(f'No risk measure found for quantity type: {quantity_type}')

        return map[quantity_type]

    def _nav_scale_orders(self, orders, first_quantity, price_measure):
        sorted_order_days = sorted(make_list(orders.keys()))
        final_days_orders = {}

        # Populate dict of dates and instruments sold on those dates
        for create_date, portfolio in orders.items():
            for inst in portfolio.all_instruments:
                d = get_final_date(inst, create_date, self.action.trade_duration)
                if d not in final_days_orders.keys():
                    final_days_orders[d] = []
                final_days_orders[d].append(inst)

        unscaled_prices_by_day = {}
        unscaled_unwind_prices_by_day = {}
        with PricingContext(is_async=True):
            for day, portfolio in orders.items():
                with PricingContext(pricing_date=day):
                    unscaled_prices_by_day[day] = portfolio.calc(price_measure)
            for unwind_day, unwind_instruments in final_days_orders.items():
                if unwind_day <= dt.date.today():
                    with PricingContext(pricing_date=unwind_day):
                        unscaled_unwind_prices_by_day[unwind_day] = Portfolio(unwind_instruments).calc(price_measure)

        # Start with first_quantity, then only use proceeds from selling instruments
        available_cash = first_quantity
        scaling_factors_by_inst = {}
        # Go through each order day of the strategy in sorted order
        for idx, cur_day in enumerate(sorted_order_days):
            scale_factor = available_cash / unscaled_prices_by_day[cur_day].aggregate()
            portfolio = orders[cur_day]
            portfolio.scale(scale_factor)
            for inst in portfolio:
                scaling_factors_by_inst[inst] = scale_factor

            available_cash = 0

            if idx + 1 < len(sorted_order_days):
                next_day = sorted_order_days[idx + 1]
            else:
                break

            # Cash received from unwinds is the cash available for the next order
            for d, p in final_days_orders.items():
                # Only consider final days between current order date and the next in an iteration
                if cur_day < d <= next_day:
                    available_cash += sum(unscaled_unwind_prices_by_day[d][inst] * scaling_factors_by_inst[inst] for
                                          inst in p)

    def _scale_order(self, orders, price_measure):
        quantity_type = self.action.trade_quantity_type
        quantity = self.action.trade_quantity

        if quantity_type == BacktestTradingQuantityType.quantity:
            for _, portfolio in orders.items():
                portfolio.scale(quantity)
        else:
            orders_risk = {}
            if quantity_type == BacktestTradingQuantityType.NAV:
                # Scale separately if strategy is NAV
                self._nav_scale_orders(orders, quantity, price_measure)
            else:
                # Scale risk daily risk to specified value otherwise
                risk_measure = EnterPositionQuantityScaledActionImpl._quantity_type_to_risk_measure(quantity_type)
                with PricingContext():
                    for day, portfolio in orders.items():
                        with PricingContext(pricing_date=day):
                            orders_risk[day] = portfolio.calc(risk_measure)

                for day, portfolio in orders.items():
                    risk_to_scale = orders_risk[day].aggregate()
                    scaling_factor = quantity / risk_to_scale
                    portfolio.scale(scaling_factor)

    def _raise_order(self,
                     state: Union[date, Iterable[date]],
                     price_measure: RiskMeasure):
        state_list = make_list(state)
        orders = {}

        with PricingContext():
            for s in state_list:
                active_portfolio = self.action.priceables
                with PricingContext(pricing_date=s):
                    orders[s] = Portfolio(active_portfolio).resolve(in_place=False)

        final_orders = {}
        for d, p in orders.items():
            new_port = []
            for t in p.result():
                t.name = f'{t.name}_{d}'
                new_port.append(t)
            final_orders[d] = Portfolio(new_port)

        self._scale_order(final_orders, price_measure)

        return final_orders

    def apply_action(self,
                     state: Union[date, Iterable[date]],
                     backtest: BackTest,
                     trigger_info: Optional[Union[EnterPositionQuantityScaledActionInfo,
                                                  Iterable[EnterPositionQuantityScaledActionInfo]]] = None):

        orders = self._raise_order(state, backtest.price_measure)

        # record entry and unwind cashflows
        for create_date, portfolio in orders.items():
            for inst in portfolio.all_instruments:
                backtest.cash_payments[create_date].append(CashPayment(inst, effective_date=create_date, direction=-1,
                                                                       scaling_parameter=self.action.trade_quantity))
                final_date = get_final_date(inst, create_date, self.action.trade_duration)
                backtest.cash_payments[final_date].append(CashPayment(inst, effective_date=final_date,
                                                                      scaling_parameter=self.action.trade_quantity))

        for s in backtest.states:
            pos = []
            for create_date, portfolio in orders.items():
                pos += [inst for inst in portfolio.instruments
                        if get_final_date(inst, create_date, self.action.trade_duration) > s >= create_date]
            if len(pos):
                backtest.portfolio_dict[s].append(pos)

        return backtest


class HedgeActionImpl(ActionHandler):
    def __init__(self, action: HedgeAction):
        super().__init__(action)

    def apply_action(self,
                     state: Union[date, Iterable[date]],
                     backtest: BackTest,
                     trigger_info: Optional[Union[HedgeActionInfo, Iterable[HedgeActionInfo]]] = None):
        state_list = make_list(state)
        if trigger_info is None or isinstance(trigger_info, HedgeActionInfo):
            trigger_info = [trigger_info for _ in range(len(state_list))]
        trigger_infos = dict(zip_longest(state_list, trigger_info))

        with HistoricalPricingContext(dates=state_list, csa_term=self.action.csa_term):
            backtest.calc_calls += 1
            backtest.calculations += len(state_list)
            f = Portfolio(self.action.priceable).resolve(in_place=False)

        for create_date, portfolio in f.result().items():
            info = trigger_infos[create_date]
            hedge_trade = portfolio.priceables[0]
            hedge_trade.name = f'{hedge_trade.name}_{create_date.strftime("%Y-%m-%d")}'
            if isinstance(hedge_trade, Portfolio):
                for instrument in hedge_trade.all_instruments:
                    instrument.name = f'{hedge_trade.name}_{instrument.name}'
            final_date = get_final_date(hedge_trade, create_date, self.action.trade_duration,
                                        self.action.holiday_calendar, info)
            active_dates = [s for s in backtest.states if create_date <= s < final_date]

            if len(active_dates):
                scaling_portfolio = ScalingPortfolio(trade=hedge_trade, dates=active_dates, risk=self.action.risk,
                                                     csa_term=self.action.csa_term,
                                                     scaling_parameter=self.action.scaling_parameter,
                                                     risk_transformation=self.action.risk_transformation)
                entry_payment = CashPayment(trade=hedge_trade, effective_date=create_date, direction=-1,
                                            scaling_parameter=self.action.scaling_parameter)
                backtest.transaction_costs[create_date] -= self.action.transaction_cost.get_cost(state, backtest,
                                                                                                 trigger_info,
                                                                                                 hedge_trade)
                exit_payment = CashPayment(trade=hedge_trade, effective_date=final_date, scale_date=create_date,
                                           scaling_parameter=self.action.scaling_parameter) \
                    if final_date <= dt.date.today() else None
                backtest.transaction_costs[final_date] -= self.action.transaction_cost.get_cost(state, backtest,
                                                                                                trigger_info,
                                                                                                hedge_trade)
                hedge = Hedge(scaling_portfolio=scaling_portfolio,
                              entry_payment=entry_payment,
                              exit_payment=exit_payment)
                backtest.hedges[create_date].append(hedge)

        return backtest


class ExitTradeActionImpl(ActionHandler):
    def __init__(self, action: ExitTradeAction):
        super().__init__(action)

    def apply_action(self,
                     state: Union[date, Iterable[date]],
                     backtest: BackTest,
                     trigger_info: Optional[Union[ExitTradeActionInfo, Iterable[ExitTradeActionInfo]]] = None):

        for s in make_list(state):

            trades_to_remove = []

            fut_dates = list(filter(lambda d: d >= s and type(d) is dt.date, backtest.states))
            for port_date in fut_dates:
                res_fut = []
                res_futures = []
                pos_fut = list(backtest.portfolio_dict[port_date].all_instruments)
                if backtest.results[port_date]:  # there are results in future dates which need removing
                    res_fut = list(backtest.results[port_date].portfolio.all_instruments)
                    res_futures = list(backtest.results[port_date].futures)

                # We expect tradable names to be defined as <ActionName>_<TradeName>_<TradeDate>
                if self.action.priceable_names:
                    # List of trade names provided -> TradeDate <= exit trigger date and TradeName is present in list
                    port_indexes_to_remove = [i for i, x in enumerate(pos_fut) if
                                              dt.datetime.strptime(x.name.split('_')[-1], '%Y-%m-%d').date() <= s and
                                              x.name.split('_')[-2] in self.action.priceable_names]
                    result_indexes_to_remove = [i for i, x in enumerate(res_fut) if
                                                dt.datetime.strptime(x.name.split('_')[-1], '%Y-%m-%d').date() <= s and
                                                x.name.split('_')[-2] in self.action.priceable_names]
                else:
                    # List of trade names not provided -> TradeDate <= exit trigger date
                    port_indexes_to_remove = [i for i, x in enumerate(pos_fut) if
                                              dt.datetime.strptime(x.name.split('_')[-1], '%Y-%m-%d').date() <= s]
                    result_indexes_to_remove = [i for i, x in enumerate(res_fut) if
                                                dt.datetime.strptime(x.name.split('_')[-1], '%Y-%m-%d').date() <= s]

                for index in sorted(port_indexes_to_remove, reverse=True):
                    # Get list of trades that have been removed to check for their future cash flow date
                    if pos_fut[index].name not in trades_to_remove:
                        trades_to_remove.append(pos_fut[index])
                    del pos_fut[index]
                for index in sorted(result_indexes_to_remove, reverse=True):
                    del res_futures[index]
                backtest.portfolio_dict[port_date] = Portfolio(tuple(pos_fut))
                if result_indexes_to_remove:
                    backtest.results[port_date] = PortfolioRiskResult(backtest.portfolio_dict[port_date],
                                                                      backtest.results[port_date].risk_measures,
                                                                      res_futures)

            for cp_date, cp_list in list(backtest.cash_payments.items()):
                if cp_date > s:
                    indexes_to_remove = [i for i, cp in enumerate(cp_list)
                                         if cp.trade.name in [x.name for x in trades_to_remove]]
                    for index in sorted(indexes_to_remove, reverse=True):
                        cp = cp_list[index]
                        prev_pos = [i for i, x in enumerate(backtest.cash_payments[s]) if cp.trade.name == x.trade.name]
                        # If trade already exists in exit trigger date cash payments, net out the position
                        if prev_pos:
                            backtest.cash_payments[s][prev_pos[0]].direction += cp.direction
                        else:
                            cp.effective_date = s
                            backtest.cash_payments[s].append(cp)
                        backtest.transaction_costs[s] -= self.action.transaction_cost.get_cost(state, backtest,
                                                                                               trigger_info, cp.trade)
                        del backtest.cash_payments[cp_date][index]
                        backtest.transaction_costs[cp_date] += self.action.transaction_cost.get_cost(state, backtest,
                                                                                                     trigger_info,
                                                                                                     cp.trade)

                    if not backtest.cash_payments[cp_date]:
                        del backtest.cash_payments[cp_date]

            for trade in trades_to_remove:
                if trade.name not in [x.trade.name for x in backtest.cash_payments[s]]:
                    backtest.cash_payments[s].append(CashPayment(trade, effective_date=s))

        return backtest


class RebalanceActionImpl(ActionHandler):
    def __init__(self, action: RebalanceAction):
        super().__init__(action)

    def apply_action(self,
                     state: Union[date, Iterable[date]],
                     backtest: BackTest,
                     trigger_info: Optional[Union[RebalanceActionInfo, Iterable[RebalanceActionInfo]]] = None):

        new_size = self.action.method(state, backtest, trigger_info)
        current_size = 0
        for trade in backtest.portfolio_dict[state]:
            if self.action.priceable.name.split('_')[-1] in trade.name:
                current_size += getattr(trade, self.action.size_parameter)
        # if we are already at the required size then do nothing.
        if new_size - current_size == 0:
            return backtest
        pos = self.action.priceable.clone(**{self.action.size_parameter: new_size - current_size,
                                             'name': f'{self.action.priceable.name}_{state}'})

        backtest.cash_payments[state].append(CashPayment(pos, effective_date=state, direction=-1,
                                                         scaling_parameter=self.action.size_parameter))
        backtest.transaction_costs[state] -= self.action.transaction_cost.get_cost(state, backtest, trigger_info, pos)
        unwind_payment = None
        cash_payment_dates = backtest.cash_payments.keys()
        for d in reversed(sorted(cash_payment_dates)):
            for cp in backtest.cash_payments[d]:
                if self.action.priceable.name.split('_')[-1] in cp.trade.name and cp.direction == 1:
                    unwind_payment = CashPayment(pos, effective_date=d, scaling_parameter=self.action.size_parameter)
                    backtest.cash_payments[d].append(unwind_payment)
                    backtest.transaction_costs[d] -= self.action.transaction_cost.get_cost(state, backtest,
                                                                                           trigger_info, pos)
                    break
            if unwind_payment:
                break

        if unwind_payment is None:
            raise ValueError("Found no final cash payment to rebalance for trade.")

        for s in backtest.states:
            if unwind_payment.effective_date > s >= state:
                backtest.portfolio_dict[s].append(pos)

        return backtest


class GenericEngineActionFactory(ActionHandlerBaseFactory):
    def __init__(self, action_impl_map=None):
        self.action_impl_map = {
            AddTradeAction: AddTradeActionImpl,
            EnterPositionQuantityScaledAction: EnterPositionQuantityScaledActionImpl,
            HedgeAction: HedgeActionImpl,
            ExitTradeAction: ExitTradeActionImpl,
            ExitAllPositionsAction: ExitTradeActionImpl,
            RebalanceAction: RebalanceActionImpl,
            AddScaledTradeAction: AddScaledTradeActionImpl,
            **(action_impl_map or {})
        }

    def get_action_handler(self, action: Action) -> ActionHandler:
        def is_eq_underlier(leg):
            if hasattr(leg, 'asset_class'):
                return isinstance(leg.asset_class, AssetClass) and leg.asset_class == AssetClass.Equity
            return leg.__class__.__name__.lower().startswith('eq')

        if isinstance(action, EnterPositionQuantityScaledAction) and \
                not all([is_eq_underlier(p) for p in action.priceables]):
            raise RuntimeError('EnterPositionQuantityScaledAction only supported for equity underliers')
        if type(action) in self.action_impl_map:
            return self.action_impl_map[type(action)](action)
        raise RuntimeError(f'Action {type(action)} not supported by engine')


class GenericEngine(BacktestBaseEngine):

    def __init__(self, action_impl_map=None, price_measure=Price):
        self.action_impl_map = {} if action_impl_map is None else action_impl_map
        self.price_measure = price_measure
        self._pricing_context_params = None
        self._initial_pricing_context = None
        self._tracing_enabled = False

    def get_action_handler(self, action: Action) -> ActionHandler:
        handler_factory = GenericEngineActionFactory(self.action_impl_map)
        return handler_factory.get_action_handler(action)

    def supports_strategy(self, strategy):
        all_actions = reduce(lambda x, y: x + y, (map(lambda x: x.actions, strategy.triggers)))
        try:
            for x in all_actions:
                self.get_action_handler(x)
        except RuntimeError:
            return False
        return True

    def new_pricing_context(self):
        """
        generate context with the same params to avoid duplication
        """
        context_params = self._pricing_context_params

        show_progress = context_params.get('show_progress', False)
        csa_term = context_params.get('csa_term')
        market_data_location = context_params.get('market_data_location')
        request_priority = context_params.get('request_priority', DEFAULT_REQUEST_PRIORITY)
        is_batch = context_params.get('is_batch', True)

        context = PricingContext(set_parameters_only=True, show_progress=show_progress, csa_term=csa_term,
                                 market_data_location=market_data_location, request_priority=request_priority,
                                 is_batch=is_batch, use_historical_diddles_only=True)

        context._max_concurrent = 1500
        context._dates_per_batch = 200

        return context

    def run_backtest(self, strategy: Strategy, start: Optional[dt.date] = None, end: Optional[dt.date] = None,
                     frequency: Optional[str] = '1m', states: Optional[Iterable[dt.date]] = None,
                     risks: Optional[Iterable[RiskMeasure]] = None, show_progress: bool = True,
                     csa_term: Optional[str] = None, visible_to_gs: bool = False, initial_value: float = 0,
                     result_ccy: Optional[Union[str, Currency]] = None, holiday_calendar: Optional[str] = None,
                     market_data_location: Optional[str] = None, is_batch: bool = True,
                     calc_risk_at_trade_exits: bool = False, pnl_explain: Optional[PnlDefinition] = None):
        """
        run the backtest following the triggers and actions defined in the strategy.  If states are entered run on
        those dates otherwise build a schedule from the start, end, frequency
        using gs_quant.datetime.relative_date.RelativeDateSchedule
        :param strategy: the strategy object
        :param start: a datetime
        :param end: a datetime
        :param frequency: str, default '1m'
        :param states: a list of dates will override the start, end, freq if provided
        :param risks: risks to run
        :param show_progress: boolean default true
        :param csa_term: the csa term to use
        :param visible_to_gs: are the contents of risk requests visible to GS (defaults to False)
        :param initial_value: initial cash value of strategy defaults to 0
        :param result_ccy: ccy of all risks, pvs and cash
        :param holiday_calendar for date maths - list of dates
        :param market_data_location: location for the market data
        :param is_batch: use websockets to reduce timeout issues
        :param calc_risk_at_trade_exits: separate results for requested risk measures on tradable exit dates;
                                         not to be included in main results but useful for PnL decomposition
        :param pnl_explain: a Pnl Definition object which defines the risk attribution and mkt data for a pnl explain
        :return: a backtest object containing the portfolios on each day and results which show all risks on all days

        """

        logger.info(f'Starting Backtest: Building Date Schedule - {dt.datetime.now()}')
        self._tracing_enabled = Tracer.get_instance().active_span is not None
        self._pricing_context_params = {'show_progress': show_progress,
                                        'csa_term': csa_term,
                                        'visible_to_gs': visible_to_gs,
                                        'market_data_location': market_data_location,
                                        'is_batch': is_batch}

        with self.new_pricing_context():
            return self.__run(strategy, start, end, frequency, states, risks, initial_value,
                              result_ccy, holiday_calendar, calc_risk_at_trade_exits, pnl_explain)

    def _trace(self, label: str):
        if self._tracing_enabled:
            return Tracer(label)
        else:
            return nullcontext()

    def __run(self, strategy, start, end, frequency, states, risks, initial_value, result_ccy, holiday_calendar,
              calc_risk_at_trade_exits, pnl_explain):
        """
        Run the backtest strategy using the ambient pricing context
        """
        with self._trace('Relative Schedule'):
            strategy_pricing_dates = RelativeDateSchedule(frequency, start, end).apply_rule(
                holiday_calendar=holiday_calendar) if states is None else states

        strategy_pricing_dates.sort()

        strategy_start_date = strategy_pricing_dates[0]
        strategy_end_date = strategy_pricing_dates[-1]

        for trigger in strategy.triggers:
            strategy_pricing_dates += [t for t in trigger.get_trigger_times()
                                       if strategy_start_date <= t <= strategy_end_date]

        strategy_pricing_dates = list(set(strategy_pricing_dates))
        strategy_pricing_dates.sort()
        pnl_risks = [] if pnl_explain is None else pnl_explain.get_risks()
        risks = list(set(make_list(risks) + strategy.risks + pnl_risks + [self.price_measure]))
        if result_ccy is not None:
            risks = [(r(currency=result_ccy) if isinstance(r, ParameterisedRiskMeasure)
                      else raiser(f'Unparameterised risk: {r}')) for r in risks]

        if result_ccy is not None:
            if isinstance(self.price_measure, ParameterisedRiskMeasure):
                price_risk = self.price_measure(currency=result_ccy)
            else:
                raiser(f'Unparameterised price measure: {self.price_measure}')
        else:
            price_risk = self.price_measure

        backtest = BackTest(strategy, strategy_pricing_dates, risks, price_risk, holiday_calendar, pnl_explain)

        logger.info('Resolving initial portfolio')
        with self._trace('Resolve initial portfolio'):
            self._resolve_initial_portfolio(strategy, backtest, strategy_start_date,
                                            strategy_pricing_dates, holiday_calendar)

        logger.info('Building simple and semi-deterministic triggers and actions')
        self._build_simple_and_semi_triggers_and_actions(strategy, backtest, strategy_pricing_dates)

        logger.info(f'Filtering strategy calculations to run from {strategy_start_date} to {strategy_end_date}')
        backtest.portfolio_dict = defaultdict(Portfolio, {k: backtest.portfolio_dict[k]
                                                          for k in backtest.portfolio_dict
                                                          if strategy_start_date <= k <= strategy_end_date})
        backtest.hedges = defaultdict(list, {k: backtest.hedges[k]
                                             for k in backtest.hedges
                                             if strategy_start_date <= k <= strategy_end_date})

        logger.info('Pricing simple and semi-deterministic triggers and actions')
        with self._trace('Pricing semi-det Triggers'):
            self._price_semi_det_triggers(backtest, risks)

        logger.info('Scaling semi-determ triggers and actions and calculating path dependent triggers and actions')
        with self._trace('Process dates') as scope:
            if scope:
                scope.span.set_tag('dates.length', len(strategy_pricing_dates))
            for d in strategy_pricing_dates:
                if scope:
                    scope.span.log_kv({'date': str(d)})
                self._process_triggers_and_actions_for_date(d, strategy, backtest, risks)

        with self._trace('Calc New Trades'):
            self._calc_new_trades(backtest, risks)

        with self._trace('Handle Cash'):
            self._handle_cash(backtest, risks, price_risk, strategy_pricing_dates, strategy_end_date, initial_value,
                              calc_risk_at_trade_exits)

        logger.info(f'Finished Backtest:- {dt.datetime.now()}')
        return backtest

    def _resolve_initial_portfolio(self, strategy, backtest, strategy_start_date, strategy_pricing_dates,
                                   holiday_calendar):
        if len(strategy.initial_portfolio):
            for index in range(len(strategy.initial_portfolio)):
                old_name = strategy.initial_portfolio[index].name
                strategy.initial_portfolio[index].name = f'{old_name}_{strategy_start_date.strftime("%Y-%m-%d")}'
                entry_payment = CashPayment(strategy.initial_portfolio[index],
                                            effective_date=strategy_start_date, direction=-1)
                backtest.cash_payments[strategy_start_date].append(entry_payment)
                final_date = get_final_date(strategy.initial_portfolio[index], strategy_start_date, None,
                                            holiday_calendar)
                exit_payment = CashPayment(strategy.initial_portfolio[index],
                                           effective_date=final_date)
                backtest.cash_payments[final_date].append(exit_payment)
            init_port = Portfolio(strategy.initial_portfolio)
            with PricingContext(strategy_start_date):
                init_port.resolve()
            for d in strategy_pricing_dates:
                backtest.portfolio_dict[d].append(init_port.instruments)

    def _build_simple_and_semi_triggers_and_actions(self, strategy, backtest, strategy_pricing_dates):
        for trigger in strategy.triggers:
            if trigger.calc_type != CalcType.path_dependent:
                triggered_dates = []
                trigger_infos = defaultdict(list)
                with self._trace('Build semi-det trigger') as scope:
                    for d in strategy_pricing_dates:
                        t_info = trigger.has_triggered(d, backtest)
                        if t_info:
                            triggered_dates.append(d)
                            if t_info.info_dict:
                                for k, v in t_info.info_dict.items():
                                    trigger_infos[k].append(v)
                    if scope:
                        scope.span.set_tag('trigger.type', type(trigger).__name__)
                        scope.span.set_tag('dates.triggered', len(triggered_dates))
                        scope.span.set_tag('action.count', len(trigger.actions))

                    for action in trigger.actions:
                        if action.calc_type != CalcType.path_dependent:
                            with self._trace('Build semi-det action') as scope:
                                if scope:
                                    scope.span.set_tag('action.type', type(action).__name__)
                                self.get_action_handler(action).apply_action(
                                    triggered_dates,
                                    backtest,
                                    trigger_infos[type(action)]
                                    if type(action) in trigger_infos else None
                                )

    def _price_semi_det_triggers(self, backtest, risks):
        with PricingContext():
            backtest.calc_calls += 1
            for day, portfolio in backtest.portfolio_dict.items():
                if isinstance(day, dt.date):
                    with PricingContext(day):
                        backtest.calculations += len(portfolio) * len(risks)
                        backtest.add_results(day, portfolio.calc(tuple(risks)))

            # semi path dependent initial calc
            for _, hedge_list in backtest.hedges.items():
                scaling_list = [h.scaling_portfolio for h in hedge_list]
                for p in scaling_list:
                    with HistoricalPricingContext(dates=p.dates):
                        backtest.calculations += len(risks) * len(p.dates)
                        port = p.trade if isinstance(p.trade, Portfolio) else Portfolio([p.trade])
                        p.results = port.calc(tuple(risks))

    def _process_triggers_and_actions_for_date(self, d, strategy, backtest: BackTest, risks):
        logger.debug(f'{d}: Processing triggers and actions')
        # path dependent
        for trigger in strategy.triggers:
            if trigger.calc_type == CalcType.path_dependent:
                if trigger.has_triggered(d, backtest):
                    for action in trigger.actions:
                        self.get_action_handler(action).apply_action(d, backtest)
            else:
                for action in trigger.actions:
                    if action.calc_type == CalcType.path_dependent:
                        if trigger.has_triggered(d, backtest):
                            self.get_action_handler(action).apply_action(d, backtest)
        # test to see if new trades have been added and calc
        port = []
        for t in backtest.portfolio_dict[d]:
            if t.name not in backtest.results[d].portfolio:
                port.append(t)

        if len(port):
            with PricingContext(pricing_date=d):
                results = Portfolio(port).calc(tuple(risks))

            backtest.add_results(d, results)

        for hedge in backtest.hedges[d]:
            sp = hedge.scaling_portfolio
            if sp.results is None:
                with HistoricalPricingContext(dates=sp.dates):
                    backtest.calculations += len(risks) * len(sp.dates)
                    port_sp = sp.trade if isinstance(sp.trade, Portfolio) else Portfolio([sp.trade])
                    sp.results = port_sp.calc(tuple(risks))

        # semi path dependent scaling
        if d in backtest.hedges:
            if len(backtest.hedges[d]) and d not in backtest.results:
                # No risk found to hedge, proceed to the next date
                return
            for hedge in backtest.hedges[d]:
                p = hedge.scaling_portfolio
                current_risk = backtest.results[d][p.risk] \
                    .transform(risk_transformation=p.risk_transformation).aggregate(allow_mismatch_risk_keys=True)
                hedge_risk = p.results[d][p.risk].transform(risk_transformation=p.risk_transformation).aggregate()
                if hedge_risk == 0:
                    continue
                if current_risk.unit != hedge_risk.unit:
                    raise RuntimeError('cannot hedge in a different currency')
                scaling_factor = current_risk / hedge_risk
                if isinstance(p.trade, Portfolio):
                    # Scale the portfolio by risk target
                    scaled_portfolio_position = copy.deepcopy(p.trade)
                    scaled_portfolio_position.name = f'Scaled_{scaled_portfolio_position.name}'
                    for instrument in scaled_portfolio_position.all_instruments:
                        instrument.name = f'Scaled_{instrument.name}'

                    # trade hedge in opposite direction
                    scale_direction = -1
                    scaled_portfolio_position.scale(scaling_factor * scale_direction)

                    for day in p.dates:
                        # add scaled hedge position to portfolio for day.
                        # NOTE this adds leaves, not the portfolio
                        backtest.portfolio_dict[day] += copy.deepcopy(scaled_portfolio_position)

                    # scale trade in hedge cash payments
                    hedge.entry_payment.trade = copy.deepcopy(scaled_portfolio_position)
                    if hedge.exit_payment is not None:
                        hedge.exit_payment.trade = copy.deepcopy(scaled_portfolio_position)
                        hedge.exit_payment.scale_date = None
                else:
                    new_notional = getattr(p.trade, p.scaling_parameter) * -scaling_factor
                    scaled_trade = p.trade.clone(**{p.scaling_parameter: new_notional, 'name': p.trade.name})
                    for day in p.dates:
                        backtest.add_results(day, p.results[day] * -scaling_factor)
                        backtest.portfolio_dict[day] += Portfolio(scaled_trade)
                # Add payments to backtest cash payments
                # Scaled if portfolio, otherwise picked up from scaled results or scaled via scale_date
                backtest.cash_payments[hedge.entry_payment.effective_date].append(hedge.entry_payment)
                if hedge.exit_payment is not None:
                    backtest.cash_payments[hedge.exit_payment.effective_date].append(hedge.exit_payment)

    def _calc_new_trades(self, backtest, risks):
        logger.info('Calculating and scaling newly added portfolio positions')
        # test to see if new trades have been added and calc
        with PricingContext():
            backtest.calc_calls += 1
            leaves_by_date = {}
            for day, portfolio in backtest.portfolio_dict.items():
                # Nothing to schedule for calculation, continue
                if not portfolio:
                    continue
                results_for_date = backtest.results[day]

                trades_for_date = results_for_date.portfolio if isinstance(results_for_date, PortfolioRiskResult) \
                    else []
                leaves = []
                for leaf in portfolio:
                    if leaf.name not in trades_for_date:
                        logger.debug(f'{day}: new portfolio position {leaf.name} scheduled for calculation')
                        leaves.append(leaf)

                if len(leaves):
                    with PricingContext(pricing_date=day):
                        leaves_by_date[day] = Portfolio(leaves).calc(tuple(risks))
                        backtest.calculations += len(leaves) * len(risks)

        logger.info('Processing results for newly added portfolio positions')
        for day, leaves in leaves_by_date.items():
            backtest.add_results(day, leaves)

    def _handle_cash(self, backtest, risks, price_risk, strategy_pricing_dates, strategy_end_date, initial_value,
                     calc_risk_at_trade_exits):
        logger.info('Calculating prices for cash payments')
        # run any additional calcs to handle cash scaling (e.g. unwinds)
        cash_results = {}
        cash_trades_by_date = defaultdict(list)
        exited_cash_trades_by_date = defaultdict(list)
        for _, cash_payments in backtest.cash_payments.items():
            for cp in cash_payments:
                # only calc if additional point is required
                trades = cp.trade.all_instruments if isinstance(cp.trade, Portfolio) else [cp.trade]
                for trade in trades:
                    if cp.effective_date and cp.effective_date <= strategy_end_date:
                        if cp.effective_date not in backtest.results or \
                                trade not in backtest.results[cp.effective_date]:
                            cash_trades_by_date[cp.effective_date].append(trade)
                            if calc_risk_at_trade_exits and cp.direction == 1:
                                exited_cash_trades_by_date[cp.effective_date].append(trade)
                        else:
                            cp.scale_date = None

        with PricingContext():
            backtest.calc_calls += 1
            for cash_date, trades in cash_trades_by_date.items():
                with PricingContext(cash_date):
                    backtest.calculations += len(risks)
                    cash_results[cash_date] = Portfolio(trades).calc(price_risk)
                    if calc_risk_at_trade_exits and cash_date in exited_cash_trades_by_date:
                        expiring_trades = exited_cash_trades_by_date[cash_date]
                        backtest.trade_exit_risk_results[cash_date] = Portfolio(expiring_trades).calc(risks)

        # handle cash
        current_value = None
        for d in sorted(set(strategy_pricing_dates + list(backtest.cash_payments.keys()))):
            if d <= strategy_end_date:
                if current_value is not None:
                    backtest.cash_dict[d] = current_value
                if d in backtest.cash_payments:
                    for cp in backtest.cash_payments[d]:
                        trades = cp.trade.all_instruments if isinstance(cp.trade, Portfolio) else [cp.trade]
                        for trade in trades:
                            value = cash_results.get(cp.effective_date, {}).get(price_risk, {}).get(trade.name, {})
                            try:
                                value = backtest.results[cp.effective_date][price_risk][trade.name] \
                                    if value == {} else value
                            except (KeyError, ValueError):
                                raise RuntimeError(f'failed to get cash value for {trade.name} on '
                                                   f'{cp.effective_date} received value of {value}')
                            if not isinstance(value, float):
                                raise RuntimeError(f'failed to get cash value for {trade.name} on '
                                                   f'{cp.effective_date} received value of {value}')
                            ccy = next(iter(value.unit))
                            if d not in backtest.cash_dict:
                                backtest.cash_dict[d] = {ccy: initial_value}
                            if ccy not in backtest.cash_dict[d]:
                                backtest.cash_dict[d][ccy] = 0
                            if cp.scale_date:
                                scale_notional = getattr(backtest.portfolio_dict[cp.scale_date][cp.trade.name],
                                                         cp.scaling_parameter)
                                scale_date_adj = scale_notional / getattr(cp.trade, cp.scaling_parameter)
                                cp.cash_paid[ccy] += value * scale_date_adj * cp.direction
                            else:
                                cp.cash_paid[ccy] += value * cp.direction

                        for ccy, cash_paid in cp.cash_paid.items():
                            backtest.cash_dict[d][ccy] += cash_paid

                    current_value = backtest.cash_dict[d]

                current_value = copy.deepcopy(current_value)



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import logging
from typing import Iterable

import gs_quant.target.backtests as backtests
from gs_quant.api.gs.backtests import GsBacktestApi
from gs_quant.api.gs.backtests_xasset.apis import GsBacktestXassetApi
from gs_quant.api.gs.backtests_xasset.request import BasicBacktestRequest
from gs_quant.api.gs.backtests_xasset.response_datatypes.backtest_datatypes import DateConfig, Trade, \
    CostPerTransaction, TransactionCostModel
from gs_quant.backtests.core import Backtest, TradeInMethod
from gs_quant.errors import MqValueError
from gs_quant.target.backtests import *
from gs_quant.instrument import EqOption, EqVarianceSwap, Instrument
from gs_quant.target.instrument import FXOption, FXBinary

_logger = logging.getLogger(__name__)

BACKTEST_TYPE_NAME = 'VolatilityFlow'
BACKTEST_TYPE_VALUE = 'Volatility Flow'
ISO_FORMAT = r"^([0-9]{4})-([0-9]{2})-([0-9]{2})$"


class StrategySystematic:
    """Equity back testing systematic strategy"""
    _supported_eq_instruments = (EqOption, EqVarianceSwap)
    _supported_fx_instruments = (FXOption, FXBinary)
    _supported_instruments = _supported_eq_instruments + _supported_fx_instruments

    def __init__(self,
                 underliers: Union[Instrument, Iterable[Instrument]],
                 quantity: float = 1,
                 quantity_type: Union[BacktestTradingQuantityType, str] = BacktestTradingQuantityType.notional,
                 trade_in_method: Union[TradeInMethod, str] = TradeInMethod.FixedRoll,
                 roll_frequency: str = None,
                 scaling_method: str = None,
                 index_initial_value: float = 0.0,
                 delta_hedge: DeltaHedgeParameters = None,
                 name: str = None,
                 cost_netting: bool = False,
                 currency: Union[Currency, str] = Currency.USD,
                 trade_in_signals: Tuple[BacktestSignalSeriesItem, ...] = None,
                 trade_out_signals: Tuple[BacktestSignalSeriesItem, ...] = None,
                 market_model: Union[EquityMarketModel, str] = EquityMarketModel.SFK,
                 roll_date_mode: str = None,
                 expiry_date_mode: str = None,
                 cash_accrual: bool = True):
        self.__cost_netting = cost_netting
        self.__currency = get_enum_value(Currency, currency)
        self.__name = name
        self.__backtest_type = BACKTEST_TYPE_NAME
        self.__cash_accrual = cash_accrual

        trade_in_method = get_enum_value(TradeInMethod, trade_in_method).value
        market_model = get_enum_value(EquityMarketModel, market_model).value

        self.__trading_parameters = BacktestTradingParameters(
            quantity=quantity,
            quantity_type=get_enum_value(BacktestTradingQuantityType, quantity_type).value,
            trade_in_method=trade_in_method,
            roll_frequency=roll_frequency,
            trade_in_signals=trade_in_signals,
            trade_out_signals=trade_out_signals,
            roll_date_mode=roll_date_mode
        )

        self.__underliers = []
        trade_instruments = []
        if isinstance(underliers, self._supported_instruments):
            instrument = underliers
            notional_percentage = 100
            instrument = self.check_underlier_fields(instrument)
            trade_instruments.append(instrument)
            self.__underliers.append(BacktestStrategyUnderlier(
                instrument=instrument,
                notional_percentage=notional_percentage,
                hedge=BacktestStrategyUnderlierHedge(risk_details=delta_hedge),
                market_model=market_model,
                expiry_date_mode=expiry_date_mode))
        else:
            for underlier in underliers:
                if isinstance(underlier, tuple):
                    instrument = underlier[0]
                    notional_percentage = underlier[1]
                else:
                    instrument = underlier
                    notional_percentage = 100

                if not isinstance(instrument, self._supported_instruments):
                    raise MqValueError('The format of the backtest asset is incorrect.')
                elif isinstance(instrument, self._supported_fx_instruments):
                    instrument = instrument.clone()
                    instrument.notional_amount *= notional_percentage / 100

                instrument = self.check_underlier_fields(instrument)
                trade_instruments.append(instrument)
                self.__underliers.append(BacktestStrategyUnderlier(
                    instrument=instrument,
                    notional_percentage=notional_percentage,
                    hedge=BacktestStrategyUnderlierHedge(risk_details=delta_hedge),
                    market_model=market_model,
                    expiry_date_mode=expiry_date_mode))
        # xasset backtesting service fields
        trade_buy_dates = tuple(s.date for s in trade_in_signals if s.value) if trade_in_signals is not None else None
        trade_exit_dates = tuple(s.date for s in trade_out_signals if s.value) if trade_out_signals is not None else \
            None
        self.__trades = (Trade(tuple(trade_instruments), roll_frequency, trade_buy_dates, roll_frequency,
                               trade_exit_dates, quantity, quantity_type),)
        self.__delta_hedge_frequency = '1b' if delta_hedge else None

        backtest_parameters_class: Base = getattr(backtests, self.__backtest_type + 'BacktestParameters')
        backtest_parameter_args = {
            'trading_parameters': self.__trading_parameters,
            'underliers': self.__underliers,
            'trade_in_method': trade_in_method,
            'scaling_method': scaling_method,
            'index_initial_value': index_initial_value
        }
        self.__backtest_parameters = backtest_parameters_class.from_dict(backtest_parameter_args)
        all_eq = all(isinstance(i, self._supported_eq_instruments) for i in trade_instruments)
        all_fx = all(isinstance(i, self._supported_fx_instruments) for i in trade_instruments)
        if not (all_eq or all_fx):
            raise MqValueError('Cannot run backtests for different asset classes.')
        self.__use_xasset_backtesting_service = all_fx

    @staticmethod
    def check_underlier_fields(
            underlier: Instrument
    ) -> Instrument:

        if isinstance(underlier, EqOption):
            underlier.number_of_options = None

        return underlier

    def __run_service_based_backtest(self, start: datetime.date, end: datetime.date,
                                     measures: Iterable[FlowVolBacktestMeasure]) -> BacktestResult:
        date_cfg = DateConfig(start, end)
        calc_measures = tuple(m for m in measures if m != FlowVolBacktestMeasure.portfolio)
        if not calc_measures:
            calc_measures = (FlowVolBacktestMeasure.PNL,)
        basic_bt_request = BasicBacktestRequest(date_cfg, self.__trades, calc_measures, self.__delta_hedge_frequency,
                                                CostPerTransaction(TransactionCostModel.Fixed, 0), None)
        basic_bt_response = GsBacktestXassetApi.calculate_basic_backtest(basic_bt_request, decode_instruments=False)
        risks = tuple(
            BacktestRisk(name=k.value,
                         timeseries=tuple(FieldValueMap(date=d, value=r.result) for d, r in v.items()))
            for k, v in basic_bt_response.measures.items()
        )
        portfolio = []
        if FlowVolBacktestMeasure.portfolio in measures:
            for d in sorted(set().union(basic_bt_response.portfolio.keys(), basic_bt_response.transactions.keys())):
                if d in basic_bt_response.portfolio:
                    positions = [{'instrument': i if i is not None else {}} for
                                 i in basic_bt_response.portfolio[d]]
                else:
                    positions = []
                transactions = []
                if d in basic_bt_response.transactions:
                    for t in basic_bt_response.transactions[d]:
                        trades = [{'instrument': i if i is not None else {},
                                   'price': t.portfolio_price}
                                  for i in t.portfolio] if t.portfolio is not None else []
                        transactions.append({'type': t.direction.value, 'trades': trades})
                portfolio.append({'date': d, 'positions': positions, 'transactions': transactions})
        return BacktestResult(risks=risks, portfolio=portfolio)

    def backtest(
            self,
            start: datetime.date = None,
            end: datetime.date = datetime.date.today() - datetime.timedelta(days=1),
            is_async: bool = False,
            measures: Iterable[FlowVolBacktestMeasure] = (FlowVolBacktestMeasure.ALL_MEASURES,),
            correlation_id: str = None
    ) -> Union[Backtest, BacktestResult]:
        if self.__use_xasset_backtesting_service:
            return self.__run_service_based_backtest(start, end, measures)
        params_dict = self.__backtest_parameters.as_dict()
        params_dict['measures'] = [m.value for m in measures]
        backtest_parameters_class: Base = getattr(backtests, self.__backtest_type + 'BacktestParameters')
        params = backtest_parameters_class.from_dict(params_dict)

        backtest = Backtest(name=self.__name,
                            mq_symbol=self.__name,
                            parameters=params,
                            start_date=start,
                            end_date=end,
                            type=BACKTEST_TYPE_VALUE,
                            asset_class=AssetClass.Equity,
                            currency=self.__currency,
                            cost_netting=self.__cost_netting,
                            cash_accrual=self.__cash_accrual)

        if is_async:
            # Create back test ...
            response = GsBacktestApi.create_backtest(backtest)

            # ... and schedule it
            GsBacktestApi.schedule_backtest(backtest_id=response.id)
        else:
            # Run on-the-fly back test
            response = GsBacktestApi.run_backtest(backtest, correlation_id)

        return response



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import pandas as pd

from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import dataclass_json, LetterCase, config

from gs_quant.api.gs.backtests_xasset.json_encoders.response_datatypes.risk_result_datatype_encoders import \
    encode_series_result, decode_series_result, encode_dataframe_result, decode_dataframe_result


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class RiskResultWithData:
    unit: Optional[str] = None

    def check_can_aggregate(self, other):
        if isinstance(other, RiskResultWithData):
            if self.unit != other.unit:
                raise ValueError(f'Cannot aggregate risk results with different units: {self.unit}, {other.unit}')
        else:
            if not isinstance(other, type(self.result)):
                raise ValueError(f'Incorrect type for other operand {type(other)}')

    def __add__(self, other):
        self.check_can_aggregate(other)
        other_operand = other.result if isinstance(other, RiskResultWithData) else other
        return type(self)(unit=self.unit, result=self.result + other_operand)

    def __radd__(self, other):
        self.check_can_aggregate(other)
        other_operand = other.result if isinstance(other, RiskResultWithData) else other
        return type(self)(unit=self.unit, result=other_operand + self.result)

    def __sub__(self, other):
        self.check_can_aggregate(other)
        other_operand = other.result if isinstance(other, RiskResultWithData) else other
        return type(self)(unit=self.unit, result=self.result - other_operand)

    def __mul__(self, other):
        self.check_can_aggregate(other)
        other_operand = other.result if isinstance(other, RiskResultWithData) else other
        return type(self)(unit=self.unit, result=self.result * other_operand)

    def __rmul__(self, other):
        self.check_can_aggregate(other)
        other_operand = other.result if isinstance(other, RiskResultWithData) else other
        return type(self)(unit=self.unit, result=other_operand * self.result)

    def __truediv__(self, other):
        self.check_can_aggregate(other)
        other_operand = other.result if isinstance(other, RiskResultWithData) else other
        return type(self)(unit=self.unit, result=self.result / other_operand)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class FloatWithData(RiskResultWithData):
    result: Optional[float] = None
    type: str = 'float'


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class StringWithData(RiskResultWithData):
    result: Optional[str] = None
    type: str = 'string'


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class VectorWithData(RiskResultWithData):
    result: Optional[pd.Series] = field(default=None, metadata=config(encoder=encode_series_result,
                                                                      decoder=decode_series_result))
    type: str = 'vector'


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class MatrixWithData(RiskResultWithData):
    result: pd.DataFrame = field(default=None, metadata=config(encoder=encode_dataframe_result,
                                                               decoder=decode_dataframe_result))
    type: str = 'matrix'



"""
Copyright 2024 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import pytest

FAILED_TESTS = []
RUN_TESTS = []
LOGGED_EVENTS = []


def did_anything_fail():
    return len(FAILED_TESTS) > 0


def did_anything_run():
    return len(RUN_TESTS) > 0


def log_mock_data_event(event: str):
    LOGGED_EVENTS.append(event)


def pytest_addoption(parser):
    parser.addoption(
        '--fixmockdata', action='store_true'
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "fixmockdata: marker to update mock file data"
    )


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_makereport(item, call) -> None:
    """Called to make report for the test for test item (the call phase).

    :param item:
        The item.
    """
    if call.when == 'call':
        RUN_TESTS.append(item.nodeid)
        if call.excinfo:
            FAILED_TESTS.append((item.nodeid, item))


def pytest_terminal_summary(terminalreporter, exitstatus, config):
    terminalreporter.ensure_newline()
    terminalreporter.section("Mock File Summary")
    if not len(LOGGED_EVENTS):
        terminalreporter.write_line("Nothing Changed")
    for event in LOGGED_EVENTS:
        terminalreporter.write_line(event)


def pytest_collection_modifyitems(items):
    LOGGED_EVENTS.clear()
    FAILED_TESTS.clear()
    for item in items:
        # Skip the integration tests unless explicitly asked for
        run_fix_mock_data = item.config.getoption("fixmockdata")  # run with pytest . --fixmockdata
        if any(marker.name == 'fixmockdata' for marker in item.own_markers) and not run_fix_mock_data:
            item.add_marker(pytest.mark.skip(reason='Skipping Mock Data updating'))



"""
Copyright 2023 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Tuple

import cachetools
import pandas as pd

from gs_quant.base import Base
from gs_quant.session import GsSession


class CacheEvent(Enum):
    PUT = 'Put'
    GET = 'Get'


class ApiRequestCache(ABC):

    def get(self, session: GsSession, key: Any, **kwargs):
        cache_lookup = self._get(session, key, **kwargs)
        if cache_lookup is not None:
            self.record(session, key, CacheEvent.GET, **kwargs)
        return cache_lookup

    @abstractmethod
    def _get(self, session: GsSession, key: Any, **kwargs):
        pass

    def record(self, session: GsSession, key: Any, method: CacheEvent, **kwargs):
        pass

    def put(self, session: GsSession, key: Any, value, **kwargs):
        self._put(session, key, value, **kwargs)
        self.record(session, key, CacheEvent.PUT, **kwargs)

    @abstractmethod
    def _put(self, session: GsSession, key: Any, value, **kwargs):
        pass


class InMemoryApiRequestCache(ApiRequestCache):

    def __init__(self, max_size=1000, ttl_in_seconds=3600):
        self._cache = cachetools.TTLCache(max_size, ttl_in_seconds)
        self._records = []

    def get_events(self) -> Tuple[Tuple[CacheEvent, Any], ...]:
        return tuple(self._records)

    def clear_events(self):
        self._records.clear()

    def _make_str_key(self, key: Any):
        if isinstance(key, (list, tuple)):
            return "_".join(self._make_str_key(k) for k in key)
        elif isinstance(key, (Base, pd.DataFrame)):
            return key.to_json()
        elif isinstance(key, dict):
            return self._make_str_key(list(key.items()))
        return str(key)

    def _get(self, session: GsSession, key: Any, **kwargs):
        return self._cache.get(self._make_str_key(key))

    def record(self, session: GsSession, key: Any, method: CacheEvent, **kwargs):
        self._records.append((method, key))

    def _put(self, session: GsSession, key, value, **kwargs):
        self._cache[self._make_str_key(key)] = value



"""
Copyright 2020 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import logging
from typing import Optional

import pandas as pd

import gs_quant.timeseries.measures_reports as ReportMeasures
from gs_quant.api.gs.data import QueryType
from gs_quant.api.gs.portfolios import GsPortfolioApi
from gs_quant.entities.entity import EntityType
from gs_quant.markets.portfolio_manager import PortfolioManager
from gs_quant.timeseries import plot_measure_entity
from gs_quant.timeseries.measures import _extract_series_from_df

LOGGER = logging.getLogger(__name__)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.AUM])
def aum(portfolio_id: str, start_date: dt.date = None, end_date: dt.date = None, *,
        source: str = None, real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Returns the Custom AUM uploaded for the portfolio
    :param portfolio_id: id of portfolio
    :param start_date: start date for getting aum
    :param end_date: end date for getting aum
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio aum values
    """
    data = GsPortfolioApi.get_custom_aum(portfolio_id, start_date, end_date)
    df = pd.DataFrame.from_records(data)
    df.set_index(pd.DatetimeIndex(df['date']), inplace=True)
    return _extract_series_from_df(df, QueryType.AUM, True)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.FACTOR_EXPOSURE])
def portfolio_factor_exposure(portfolio_id: str, risk_model_id: str, factor_name: str, unit: str = 'Notional',
                              benchmark_id: str = None, *, source: str = None,
                              real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor exposure data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name
    :param unit: unit in which the timeseries is returned (Notional or Percent)
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor exposure for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.factor_exposure(report.id, factor_name, unit)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.FACTOR_PNL])
def portfolio_factor_pnl(portfolio_id: str, risk_model_id: str, factor_name: str, unit: str = 'Notional',
                         benchmark_id: str = None, *, source: str = None,
                         real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor PnL data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name
    :param unit: unit in which the timeseries is returned (Notional or Percent)
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor pnl for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.factor_pnl(report.id, factor_name, unit)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.FACTOR_PROPORTION_OF_RISK])
def portfolio_factor_proportion_of_risk(portfolio_id: str, risk_model_id: str, factor_name: str,
                                        benchmark_id: str = None, *, source: str = None,
                                        real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor proportion of risk data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor proportion of risk for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.factor_proportion_of_risk(report.id, factor_name)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.DAILY_RISK])
def portfolio_daily_risk(portfolio_id: str, risk_model_id: str, factor_name: str = 'Total',
                         benchmark_id: str = None, *, source: str = None,
                         real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Daily risk data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name (must be "Factor", "Specific", or "Total")
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily risk for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.daily_risk(report.id, factor_name)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.ANNUAL_RISK])
def portfolio_annual_risk(portfolio_id: str, risk_model_id: str, factor_name: str = 'Total',
                          benchmark_id: str = None, *, source: str = None,
                          real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Annual risk data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name (must be "Factor", "Specific", or "Total")
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily risk for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.annual_risk(report.id, factor_name)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.THEMATIC_EXPOSURE])
def portfolio_thematic_exposure(portfolio_id: str, basket_ticker: str, *, source: str = None,
                                real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Thematic exposure of a portfolio to a requested GS thematic flagship basket

    :param portfolio_id: portfolio id
    :param basket_ticker: ticker for thematic basket
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily thematic beta of portfolio to requested flagship basket
    """
    pm = PortfolioManager(portfolio_id)
    thematic_report = pm.get_thematic_report()
    return ReportMeasures.thematic_exposure(thematic_report.id, basket_ticker)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.PNL])
def portfolio_pnl(portfolio_id: str, unit: str = 'Notional', *, source: str = None,
                  real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    PNL from all holdings, if unit is Percent, geometrically aggregate over time frame

    :param portfolio_id: id of portfolio
    :param unit: by default uses mode as Notional, but can also be set to Percent
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio pnl
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.pnl(performance_report.id, unit)
